<?php
/**
 * DyaPress e-Commerce Ajax Handler.
 *
 * @class    DyaPressECommerceAjaxHandler
 * @package  DyaPressECommerce/Classes
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * WordPress DyaPress e-Commerce plugin :
 * Copyright (C) 2018-2020      François Pons        <fpons@aytechnet.fr>
 *
 * Please see https://dyapress.com/
 */

defined( 'ABSPATH' ) or die( 'Oops' );

if ( !class_exists( 'DyaPressECommerceAjaxHandler' ) ) {
    class DyaPressECommerceAjaxHandler {
	public static function init() {
	    $handler = array(
		'process_order',
		'get_product_infos',
		'is_vat_applicable',
		'reset_password',
		'registration',
		'registration_purchase',
		'confirm_purchase',
	    );
	    $handler_priv = array(
		'get_account_section',
		'account_update',
	    );

	    foreach ($handler as $name ) {
		add_action( 'wp_ajax_dyapress_' . $name, __CLASS__ . '::' . $name );
		add_action( 'wp_ajax_nopriv_dyapress_' . $name, __CLASS__ . '::' . $name );
	    }
	    foreach ($handler_priv as $name ) {
		add_action( 'wp_ajax_dyapress_' . $name, __CLASS__ . '::' . $name );
	    }

	    add_action( 'wp_enqueue_scripts', __CLASS__ . '::add_js_scripts' );
	}

	public static function add_js_scripts() {
	    $dyapress = array(
		'ajax_url' => admin_url( 'admin-ajax.php' ),
		'european_countries' => dyapress_premium()->get_european_countries(),
		'stripe_public_key' => dyapress_premium()->stripe_public_key(),
		'current_user' => dyapress_premium()->get_current_user(),
		'default_currency' => dyapress_premium()->default_currency(),
		'messages' => array(
		    'ok' => __( 'Your order has been validated.', 'dyapress-ecommerce' ),
		    'update_thirdparty' => __( 'Update thirdparty', 'dyapress-ecommerce' ),
		    'add_thirdparty' => __( 'Add new thirdparty', 'dyapress-ecommerce' ),
		    'processing' => __( 'Processing&hellip;', 'dyapress-ecommerce' ),
		    'processing_order' => __( 'Your order is being processed.', 'dyapress-ecommerce' ),
		    'processing_update' => __( 'Processing update', 'dyapress-ecommerce' ),
		    'card_action_failed' => __( 'Payment authentification failed and your order is not validated! Please try again!', 'dyapress-ecommerce' ),
		),
	    );

	    wp_enqueue_script( 'dyapress-ajax', plugins_url( 'dyapress-ecommerce/js/ajax-handler.js' ), array( 'jquery' ) );
	    wp_localize_script( 'dyapress-ajax', 'dyapress_ecommerce', apply_filters( 'dyapress_ajax_js_init', $dyapress ) );
	}

	public static function process_order() {
	    if ( isset( $_POST['order'] ) && isset( $_POST['premium_uuid'] ) && isset( $_POST['order_secret'] ) ) {
		$order = dyapress_premium()->sanitize_order( stripslashes( $_POST['order'] ), $_POST['premium_uuid'], $_POST['order_secret'] );

		if ( !empty( $order ) ) {
		    if ( isset( $_POST['ordered_products'] ) )
			$ordered_products = json_decode( stripslashes( $_POST['ordered_products'] ) );

		    // ordered_products is a map of up-to-date product data so update internal cache
		    // FIXME: update code below to something like : dyapress_premium()->set_order_transient( $order );
		    if ( !empty( $ordered_products ) ) {
			$uuid = dyapress_premium()->uuid();

			foreach ( $ordered_products as $id => $product ) {
			    dyapress_premium()->set_transient_product( $product, $id, $uuid );
			}
		    }

		    if ( !empty( $order->order_meta ) && !empty( $order->order_meta->order_name ) )
			do_action( 'dyapress_premium_process_order_' . $order->order_meta->order_name, $order, $ordered_products );

		    do_action( 'dyapress_premium_process_order', $order, $ordered_products );
		}
	    }

	    die();
	}

	public static function get_product_infos() {
	    if ( !empty( $_POST['product_id'] ) && !empty( $_POST['product_quantity'] ) ) {
		$product_id = dyapress_premium()->map_uuid( $_POST['product_id'] );
		$product_quantity = (float)$_POST['product_quantity'];

		$result = array(
		    'product_price' => dyapress_premium()->get_product_price( $product_id, $product_quantity ),
		    'product_vat' => dyapress_premium()->get_product_vat( $product_id ),
		);

		print json_encode( $result );
	    }

	    die();
	}

	public static function is_vat_applicable() {
	    $thirdparty = array();

	    if ( !empty( $_POST['country_id'] ) )
		$thirdparty['country_id'] = (int)$_POST['country_id'];
	    else
		$thirdparty['country_id'] = 0;

	    if ( !empty( $_POST['vat_number'] ) )
		$thirdparty['vat_number'] = sanitize_text_field( $_POST['country_id'] );
	    else
		$thirdparty['vat_number'] = '';

	    $result = 1;
	    if ( !dyapress_premium()->is_vat_applicable( $thirdparty ) )
		$result = 0;

	    print json_encode( $result );

	    die();
	}

	public static function get_account_section() {
	    $name = $_POST['name'];

	    if ( $name != '' ) {
		foreach (dyapress_ecommerce()->get_account_sections() as $section) {
		    if ( isset( $section['name'] ) && $section['name'] == $name ) {
			if ( isset( $section['content'] ) ) {
			    if ( is_callable( $section['content'] ) ) {
				ob_start();

				call_user_func( $section['content'] );

				$result = ob_get_clean();
			    } elseif ( !empty( $section['content'] ) )
				$result = $section['content'];
			}
		    }
		}
	    }

	    print json_encode( $result );

	    die();
	}

	public static function account_update() {
	    $param = $_POST['param'];

	    if ( isset( $param['nonce']) && wp_verify_nonce( $param['nonce'], 'account_update' ) ) {
		$message = '';

		if ( !empty( $param['password'] ) && !empty( $param['password2'] ) ) {
		    if ( $param['password'] == $param['password2'] ) {
			$user = wp_get_current_user();
			self::reset_password( $user, $param['password'] );

			$message = __( 'Your personnal informations and password have been updated successfully', 'dyapress-ecommerce' );
		    } else {
			print json_encode( apply_filters( 'dyapress_ecommerce_account_update', array(
			    'failed' => true,
			    'message' => '<div class="failed">'.__( 'The passwords doesn\'t match so no modification have been made.', 'dyapress-ecommerce' ).'</div>',
			) ) );

			die();
		    }
		} else
		    $message = __( 'Your personnal informations have been updated successfully.', 'dyapress-ecommerce' );

		$email = dyapress_premium()->update_user( dyapress_premium()->get_current_user(), $param );
		if ( isset( $email ) ) {
		    if ( !empty( $email ) )
			$message .= '&nbsp;'.__( 'You have modified your email, you will have to activate your new email but this is not yet supported...', 'dyapress-ecommerce' );

		    print json_encode( apply_filters( 'dyapress_ecommerce_account_update', array(
			'success' => true,
			'message' => '<div class="success">'.$message.'</div>',
			'email' => $email,
			) ) );
		} else
		    print json_encode( apply_filters( 'dyapress_ecommerce_account_update', array(
			'failed' => true,
			'message' => '<div class="failed">'.__( 'An error occured when updating your account.', 'dyapress-ecommerce' ).'</div>',
		    ) ) );
	    } else
		print json_encode( apply_filters( 'dyapress_ecommerce_account_update', array(
		    'failed' => true,
		    'message' => '<div class="failed">'.__( 'You are not allowed to update your account.', 'dyapress-ecommerce' ).'</div>',
		) ) );

	    die();
	}

	public static function reset_password() {
	    $param = $_POST['param'];

	    if ( isset( $param['nonce']) && wp_verify_nonce( $param['nonce'], 'reset_password' ) ) {
		if ( isset( $param['password'] ) && isset( $param['password2'] ) && $param['password'] == $param['password2'] &&
		    isset( $param['user_id'] ) && isset( $param['email'] ) &&
		    (int)$param['user_id'] == email_exists( $param['email']) ) {
		    $user = new WP_User( (int)$param['user_id'] );
		    reset_password( $user, $param['password'] );

		    print json_encode( apply_filters( 'dyapress_ecommerce_reset_password', array(
			'success' => true,
			'message' => '<div class="success">'.__( 'Your password has been updated successfully.', 'dyapress-ecommerce' ).'</div>',
		    ) ) );
		} else {
		    print json_encode( apply_filters( 'dyapress_ecommerce_reset_password', array(
			'failed' => true,
			'message' => '<div class="failed">'.__( 'Your password has not been updated since they both mismatchs.', 'dyapress-ecommerce' ),
		    ) ) );
		}
	    } else {
		print json_encode( apply_filters( 'dyapress_ecommerce_reset_password', array(
		    'failed' => true,
		    'message' => '<div class="failed">'.__( 'You are not allowed to update your password.', 'dyapress-ecommerce' ).'</div>',
		) ) );
	    }

	    die();
	}

	public static function registration() {
	    $param = stripslashes_deep( $_POST['param'] );

	    $user = dyapress_premium()->get_user( $param );

	    if ( !isset( $user ) )
		print json_encode( apply_filters( 'dyapress_ecommerce_registration', array(
		    'failed' => true,
		    'message' => dyapress_ecommerce()->get_template_part( 'errors/user_already_exists' ),
		) ) );
	    elseif ( isset( $param['nonce']) && wp_verify_nonce( $param['nonce'], 'registration' ) ) {
		if ( empty( $user ) ) {
		    // user is not registered, we need to send an activation link
		    // which contains enough data to confirm_order with the user created or chosen from DyaPress/Dolibarr
		    dyapress_ecommerce()->send_mail(
			$param['email'], 
			sprintf( __( 'Welcome to %s! Please activate your account', 'dyapress-ecommerce' ), get_bloginfo( 'name' ), $order->id ),
			'welcome',
			$param
		    );
		}

		// the payment may have been made immediately without needing a confirmation
		// or no payment is made and should be made later, but the order is valid
		print json_encode( apply_filters( 'dyapress_ecommerce_registration', array(
		    'success' => true,
		    'user' => $user,
		    'param' => $param,
		    'message' => dyapress_ecommerce()->get_template_part(
			'registration_confirmed',
			!empty( $param['template_name'] ) ? $param['template_name'] : '',
			array_merge( $param, array( 'user' => $user ) )
		    ),
		) ) );
	    } else
		print json_encode( apply_filters( 'dyapress_ecommerce_registration', array(
		    'failed' => true,
		    'message' => dyapress_ecommerce()->get_template_part( 'errors/nonce_failed' ),
		) ) );

	    die();
	}

	public static function registration_purchase() {
	    $param = stripslashes_deep( $_POST['param'] );

	    // check if user is already identified but match current_user, or if user is not registered
	    // get_user() will return the user found or false if not found, and null if user already exists but is different
	    //            in case of the user is already logged in and email match logged in email and a thirdparty already exists (so email has already be validated)
	    //            then thirdparty_id will be defined only if thirdparty is found
	    $user = dyapress_premium()->get_user( $param );

	    if ( !isset( $user ) )
		print json_encode( apply_filters( 'dyapress_ecommerce_registration_purchase', array(
		    'failed' => true,
		    'message' => dyapress_ecommerce()->get_template_part( 'errors/user_already_exists' ),
		) ) );
	    elseif ( isset( $param['nonce']) && wp_verify_nonce( $param['nonce'], 'registration_purchase' ) ) {
		$lines = array();

		foreach ( $param as $name => $val ) {
		    if ( preg_match( '/product_quantity\[(\d+)/', $name, $matches ) ) {
			$product_id = intval($matches[1]);
			$product_quantity = intval($val);

			if ( isset( $param['product_price['.$product_id.']'] ) )
			    $product_price = $param['product_price['.$product_id.']'];
			else
			    $product_price = dyapress_premium()->get_product_price( $product_id, $product_quantity );

			if ( isset( $param['product_vat['.$product_id.']'] ) )
			    $product_vat = $param['product_vat['.$product_id.']'];
			else
			    $product_vat = dyapress_premium()->get_product_vat( $product_id, $user ? $user : $param );

			if ( isset( $param['product_reduction['.$product_id.']'] ) )
			    $product_reduction = $param['product_reduction['.$product_id.']'];
			elseif ( isset( $param['product_reduction_percent['.$product_id.']'] ) )
			    $product_reduction = $product_price * $param['product_reduction_percent['.$product_id.']'] / 100;
			else
			    $product_reduction = 0;

			$lines[] = array(
			    'order_product_id' => $product_id,
			    'order_quantity' => $product_quantity,
			    'order_price' => $product_price,
			    'order_reduction' => $product_reduction,
			    'order_vat' => $product_vat
			);
		    }
		}

		if ( count($lines) > 0 ) {
		    $meta = array();

		    if ( !empty( $param['currency'] ) )
			$currency = $param['currency'];
		    else
			$currency = null;
		    if ( !empty( $param['order_name'] ) )
			$meta['order_name'] = $param['order_name'];

		    $cookie_name = 'dyapress_ecommerce_order_';
		    if ( defined( 'COOKIEHASH' ) )
			$cookie_name .= COOKIEHASH;

		    if ( isset( $_COOKIE[$cookie_name] ) ) {
			$order = dyapress_premium()->get_order( dyapress_premium()->map_uuid( $_COOKIE[$cookie_name] ) );

			// if order exists and is not already confirmed or paid
			if ( empty( $order ) || !is_object( $order ) || $order->thirdparty_id > 0 || $order->total_capturable > 0 )
			    $order = null;
		    }
		    if ( empty( $order ) )
			$order = apply_filters( 'dyapress_ecommerce_order_created', dyapress_premium()->create_order( $lines, $currency, $meta ), $param );
		    else
			$order = apply_filters( 'dyapress_ecommerce_order_updated', dyapress_premium()->update_order( $order, $lines, $currency, $meta ), $order, $param );

		    // save back cookie with 90 days duration
		    setcookie( $cookie_name, dyapress_premium()->set_map_uuid( $_COOKIE[$cookie_name], $order->id ), time() + 60*60*24*90, COOKIEPATH ? COOKIEPATH : '/', COOKIE_DOMAIN );

		    if ( $order ) {
			// try payment if possible, Stripe is built-in and will be tried if applicable
			// payment will be false if no payment habe been made but this is not an error, null on error
			// payment will be an object if a payment has been possible (but may need to be confirmed)
			$payment = apply_filters( 'dyapress_ecommerce_payment_created', dyapress_premium()->create_order_payment( $order, $param ), $param );

			if ( isset( $payment ) ) {
			    if ( $payment ) {
				if ( $payment->requires_confirmation ) {
				    // payment need a confirmation from front-end script (or an addition action from the user)
				    // a nonce is created now for the coming order confirmation call
				    // the order confirmation will cause the payment to be confirmed
				    print json_encode( apply_filters( 'dyapress_ecommerce_registration_purchase', array(
					'success' => true,
					'order' => $order,
					'payment' => $payment,
					'user' => $user,
					'param' => $param,
					'nonce' => wp_create_nonce( 'confirm_purchase' ),
				    ) ) );

				    die();
				}
			    }

			    // payment is validated without needing confirmation so the order can be confirmed immediately
			    // and assigned to the user, the order will be no longer be modifiable
			    self::return_purchase_confirmed( $order, $payment, $user, $param );

			    die();
			}
		    }
		}

		print json_encode( apply_filters( 'dyapress_ecommerce_registration_purchase', array(
		    'failed' => true,
		    'order' => $order,
		    'user' => $user,
		    'message' => dyapress_ecommerce()->get_template_part(
			'errors/purchase_failed',
			$param['template_name'] ? $param['template_name'] : '',
			array_merge( $param, array( 'order' => $order ) ) ),
		) ) );
	    } else
		print json_encode( apply_filters( 'dyapress_ecommerce_registration_purchase', array(
		    'failed' => true,
		    'message' => dyapress_ecommerce()->get_template_part( 'errors/nonce_failed' ),
		) ) );

	    die();
	}

	public static function confirm_purchase() {
	    $param = $_POST['param'];

	    if ( isset( $param['nonce'] ) && wp_verify_nonce( $param['nonce'], 'confirm_purchase' ) ) {
		$payment = dyapress_premium()->confirm_payment( $param['payment'] );

		if ( $payment )
		    self::return_purchase_confirmed( dyapress_premium()->get_order( $param['order'] ), $payment, $param['user'], $param['param'] );
		else
		    print json_encode( apply_filters( 'dyapress_ecommerce_registration_purchase', array(
			'failed' => true,
			'order' => $param['order'],
			'user' => $param['user'],
			'message' => dyapress_ecommerce()->get_template_part(
			    'errors/purchase_failed',
			    $param['template_name'] ? $param['template_name'] : '',
			    array_merge( $param, array( 'order' => $order, 'user' => $user ) ) ),
		    ) ) );
	    } else
		print json_encode( apply_filters( 'dyapress_ecommerce_registration_purchase', array(
		    'failed' => true,
		    'message' => dyapress_ecommerce()->get_template_part( 'errors/nonce_failed' ),
		) ) );

	    die();
	}

	private static function return_purchase_confirmed( $order, $payment, $user, $param ) {
	    // if the order is effectively confirmed (since payment is confirmed and order is paid)
	    if ( $order_updated = dyapress_premium()->confirm_order( $order, $user ) ) {
		$order = $order_updated;

		// send a purchased confirmation email alone (without needed an activation)
		dyapress_ecommerce()->send_mail(
		    $user, 
		    sprintf( __( 'Your order #%d on %s is confirmed', 'dyapress-ecommerce' ), $order->id, get_bloginfo( 'name' ) ),
		    'purchase_confirmed',
		    array_merge( $param, array( 'order' => $order, 'payment' => $payment, 'user' => $user ) )
		);
	    } else {
		// user is not registered, we need to send an activation link
		// which contains enough data to confirm_order with the user created or chosen from DyaPress/Dolibarr

		// send purchase confirmation mail with activation link if needed
		dyapress_ecommerce()->send_mail(
		    $param['email'], 
		    sprintf( __( 'Welcome to %s! Please activate your account to manage your order #%d', 'dyapress-ecommerce' ), get_bloginfo( 'name' ), $order->id ),
		    'welcome_purchase_confirmed',
		    array_merge( $param, array( 'order' => $order, 'payment' => $payment ) )
		);
	    }

	    // the payment may have been made immediately without needing a confirmation
	    // or no payment is made and should be made later, but the order is valid
	    print json_encode( apply_filters( 'dyapress_ecommerce_registration_purchase', array(
		'success' => true,
		'order' => $order,
		'payment' => $payment,
		'user' => $user,
		'param' => $param,
		'message' => dyapress_ecommerce()->get_template_part(
		    'purchase_confirmed',
		    !empty( $param['template_name'] ) ? $param['template_name'] : '',
		    array_merge( $param, array( 'order' => $order, 'payment' => $payment, 'user' => $user ) )
		),
	    ) ) );
	}

    }

    DyaPressECommerceAjaxHandler::init();
}
