<?php
/**
 * DyaPress e-Commerce.
 *
 * @package DyaPressECommerce
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * WordPress DyaPress e-Commerce plugin :
 * Copyright (C) 2018-2020      François Pons        <fpons@aytechnet.fr>
 *
 * Please see https://dyapress.com/
 */

defined( 'ABSPATH' ) && defined( 'DYAPRESS_ECOMMERCE_PLUGIN_PATH' ) or die( 'Oops' );

/**
 * Main DyaPress e-Commerce Class.
 *
 * @class DyaPressECommerce
 */
final class DyaPressECommerce {
    private static $instance = null;

    private $options;
    private $settings_tab = array();
    private $account_sections = array();
    private $order_state = array();

    /**
     * Main DyaPress e-Commerce Instance.
     *
     * Ensures only one instance of DyaPress Premium e-Commerce is loaded or can be loaded.
     *
     * @static
     * @return DyaPressECommerce - Main instance.
     */
    static function get_instance() {
	if ( is_null( self::$instance ) ) {
	    self::$instance = new DyaPressECommerce();
	}

	return self::$instance;
    }

    /**
     * DyaPressECommerce Constructor.
     */
    private function __construct() {
	$this->options = get_option( 'dyapress_ecommerce_options' );

	add_action( 'init', array( &$this, 'init' ) );
	add_action( 'wp_loaded', array( &$this, 'init_download' ) );
	add_action( 'admin_init', array( &$this, 'register_options' ) );
	add_action( 'admin_menu', array( &$this, 'admin_add_pages' ) );
	add_action( 'admin_enqueue_scripts', array( &$this, 'admin_enqueue_scripts' ) );
	add_action( 'admin_post_dyapress_ecommerce_update_options', array( &$this, 'admin_update_options' ) );
	add_action( 'wp_enqueue_scripts', array( &$this, 'enqueue_scripts' ) );
	add_filter( 'display_post_states', array( &$this, 'display_post_states' ), 10, 2 );
	add_filter( 'the_content', array( &$this, 'myaccount_content' ) );

	add_filter( 'dyapress_premium_create_order_payment', array( &$this, 'filter_create_order_payment' ) );

	$this->register_order_state( "draft", __( 'Draft', 'dyapress-ecommerce' ), "#e5e5e5", "#777" );
	$this->register_order_state( "confirmed", __( 'Confirmed', 'dyapress-ecommerce' ), "#ddeeff", "#777" );
	$this->register_order_state( "pending", __( 'Pending', 'dyapress-ecommerce' ), "#4060e0", "white" );
	$this->register_order_state( "processing", __( 'Processing', 'dyapress-ecommerce' ), "#ffa020", "white" );
	$this->register_order_state( "preparing", __( 'Preparing', 'dyapress-ecommerce' ), "#ff8000", "white" );
	$this->register_order_state( "backorder", __( 'Backorder', 'dyapress-ecommerce' ), "#ff6030", "white" );
	$this->register_order_state( "completing", __( 'Completing', 'dyapress-ecommerce' ), "#20c020", "white" );
	$this->register_order_state( "completed", __( 'Completed', 'dyapress-ecommerce' ), "#108010", "white" );
	$this->register_order_state( "interrupted", __( 'Interrupted', 'dyapress-ecommerce' ), "#ff40b0", "white" );
	$this->register_order_state( "cancelled", __( 'Cancelled', 'dyapress-ecommerce' ), "#f01040", "white" );
    }

    /**
     * Init when WordPress Initialises.
     */
    function init() {
	load_plugin_textdomain( 'dyapress-ecommerce', false, 'dyapress-ecommerce/languages/' );

	require_once DYAPRESS_ECOMMERCE_PLUGIN_PATH . '/includes/class-ajax-handler.php';
	require_once DYAPRESS_ECOMMERCE_PLUGIN_PATH . '/includes/class-shortcodes.php';

	add_filter( 'set-screen-option', array( &$this, 'admin_page_set_options' ), 11, 3 );

	do_action( 'dyapress_ecommerce_init' );
    }

    /**
     * Init download when WordPress is fully loaded.
     */
    public function init_download() {
	require_once DYAPRESS_ECOMMERCE_PLUGIN_PATH . '/includes/class-download.php';
    }

    /**
     * Init administration for WordPress admin.
     */
    public function admin_add_pages() {
	require_once DYAPRESS_ECOMMERCE_PLUGIN_PATH . '/includes/class-admin.php';

	DyaPressECommerceAdmin::add_pages();
    }

    /**
     * Init administration for WordPress admin update of options.
     */
    public function admin_update_options() {
	require_once DYAPRESS_ECOMMERCE_PLUGIN_PATH . '/includes/class-admin.php';

	DyaPressECommerceAdmin::update_options();
    }

    /**
     * Init administration options for WordPress admin orders page.
     */
    public function admin_page_set_options( $status, $option, $value ) {
	if ( 'dyapress_ecommerce_orders_per_page' == $option )
	    return (int)$value;

	return $status;
    }

    /**
     * Create the DyaPress Premium menu page.
     *
     * @param string $page_title Page title as used by add_menu_page().
     * @param string $menu_title Menu title as used by add_menu_page().
     * @param string $capability Capability as used by add_menu_page().
     * @param callable $function Callable function as used by add_menu_page().
     *
     * @return string The DyaPress Premium page's hook_suffix.
     */
    public function add_menu_page( $page_title, $menu_title, $capability, $function ) {
	return add_menu_page( $page_title, $menu_title, $capability, 'dyapress-ecommerce', $function, plugins_url( '/dyapress-ecommerce/images/menu-icon'.( isset( $_GET['page'] ) && preg_match( '/^dyapress/i', $_GET['page'] ) ? '-active' : '' ).'.png' ), 55 );
    }

    /**
     * Create a DyaPress Premium submenu page.
     *
     * @api
     *
     * @param string $page_title Page title as used by add_submenu_page().
     * @param string $menu_title Menu title as used by add_submenu_page().
     * @param string $capability Capability as used by add_submenu_page().
     * @param string $menu_slug Menu slug as used by add_submenu_page().
     * @param callable $function Callable function as used by add_submenu_page().
     * @param float $position Position as used by add_submenu_page().
     *
     * @return string|false The DyaPress Premium subpage's hook_suffix, or false if the user does not have the capability required.
     */
    public function add_submenu_page( $page_title, $menu_title, $capability, $menu_slug = '', $function = '', $position = null ) {
	return add_submenu_page( 'dyapress-ecommerce', $page_title, $menu_title, $capability, ( empty( $menu_slug ) ? 'dyapress-ecommerce' : ( preg_match( '/_/', $menu_slug ) ? $menu_slug : 'dyapress_ecommerce_'.$menu_slug ) ), $function, $position );
    }

    /**
     * Retieves the URL to the given admin page and tab for DyaPress Premium.
     *
     * @api
     *
     * @param string $page Menu slug for the specific page, or empty for the main page of DyaPress Premium.
     * @param string $tab Tab or empty for no specifi tab.
     *
     * @return string Admin URL link to the DyaPress Premium page and tab.
     */
    public function admin_url( $page = false, $tab = false ) {
	$admin_url = admin_url( 'admin.php?page='.( !empty( $page ) ? ( preg_match( '/_/', $page ) ? $page : 'dyapress_ecommerce_'.$page ) : 'dyapress-ecommerce' ) );

	if ( !empty( $tab) )
	    $admin_url .= '&tab='.$tab;

	return $admin_url;
    }

    /**
     * Add a DyaPress Premium settings tab.
     *
     * @api
     *
     * @param string $title Title of the tab.
     * @param string $slug Slug of the tab.
     * @param callable $function The function to be called to output the content of this tab.
     * @param float $position The position order this tab should appear.
     */
    public function add_settings_tab( $title, $slug, $function, $position ) {
	$this->settings_tabs[$slug] = array(
	    'title' => $title,
	    'slug' => $slug,
	    'function' => $function,
	    'position' => $position,
	);
    }

    /**
     * Get all tabs sorted by position in ascending order.
     *
     * @return array Tabs array sorted by position.
     */
    public function get_settings_tabs() {
	uasort( $this->settings_tabs, function( $a, $b ) {
	    return $a['position'] - $b['position'];
	} );

	return $this->settings_tabs;
    }

    public function admin_order_columns( $args = null, $url = null ) {
	if ( isset( $args ) && is_array( $args ) && !empty( $url ) ) {
	    if ( !empty( $args['orderby' ] ) )
		$orderby = sanitize_key( $args['orderby'] );
	    else
		$orderby = 'id';

	    if ( !empty( $args['order' ] ) )
		$order = strtolower( $args['order'] );
	    if ( empty( $order ) || ( $order != 'desc' && $order != 'asc' ) )
		$order = $orderby == 'id' || $orderby == 'date' ? 'desc' : 'asc';

	    $sort_order_id = $orderby == 'id' ? ' sorted '.$order : ' sortable desc';
	    $order_id_url = $url.( !empty( $args['order_state'] ) ? '&order_state='.sanitize_key( $args['order_state'] ) : '' ).'&orderby=id&order='.( $orderby == 'id' && $order == 'desc' ? 'asc' : 'desc' );
	}

	return apply_filters( 'dyapress_ecommerce_admin_order_columns', array(
	    'cb' => array(
		'tag' => 'td',
		'class' => 'manage-column column-cb check-column',
		'content' => '<label class="screen-reader-text" for="cb-select-all-1">'.__( 'Select all', 'dyapress-ecommerce' ).'</label><input id="db-select-all-1" type="checkbox">',
	    ),
	    'id' => array(
		'class' => 'manage-column column-id column-primary'.( isset( $sort_order_id ) ? $sort_order_id : '' ),
		'content' => ( isset( $order_id_url ) ? '<a href="'.esc_url( $order_id_url ).'">' : '' ).'<span>'.__( 'ID', 'dyapress-ecommerce' ).'</span>'.( isset( $order_id_url ) ? '<span class="sorting-indicator"></span></a>' : '' ),
	    ),
	    'user' => array(
		'class' => 'manage-column column-user',
		'content' => __( 'User', 'dyapress-ecommerce' ),
	    ),
	    'thirdparty' => array(
		'class' => 'manage-column column-thirdparty',
		'content' => __( 'Thirdparty', 'dyapress-ecommerce' ),
	    ),
	    'total' => array(
		'class' => 'manage-column column-total',
		'content' => __( 'Total', 'dyapress-ecommerce' ),
	    ),
	    'capturable' => array(
		'class' => 'manage-column column-capturable',
		'content' => __( 'Capturable', 'dyapress-ecommerce' ),
	    ),
	    'paid' => array(
		'class' => 'manage-column column-paid column-price',
		'content' => __( 'Paid', 'dyapress-ecommerce' ),
	    ),
	    'payment' => array(
		'class' => 'manage-column column-payment',
		'content' => __( 'Payment', 'dyapress-ecommerce' ),
	    ),
	    'status' => array(
		'class' => 'manage-column column-status column-order_state',
		'content' => __( 'Status', 'dyapress-ecommerce' ),
	    ),
	    'date_created' => array(
		'class' => 'manage-column column-date column-date_created',
		'content' => __( 'Date', 'dyapress-ecommerce' ),
	    ),
	    'order' => array(
		'class' => 'manage-column column-order column-pdf',
		'content' => __( 'Order', 'dyapress-ecommerce' ),
	    ),
	    'invoice' => array(
		'class' => 'manage-column column-invoice column-pdf column-primary',
		'content' => __( 'Invoice', 'dyapress-ecommerce' ),
	    ),
	) );
    }

    public function admin_order_display( $o, $date_format = null ) {
	if ( empty( $date_format ) )
	    $date_format = apply_filters( 'dyapress_ecommerce_admin_date_format', 'd/m/Y H:i:s' );

	$payment_types = array();
	foreach ( dyapress_premium()->get_payment_types() as $payment ) {
	    $payment_types[$payment->id] = $payment->name;
	}

	$p = array(
	    'cb' => array(
		'tag' => 'th',
		'class' => 'check-column',
		'content' => '<label class="screen-reader-text" for="cb-select-'.$o->id.'">'.sprintf( __( 'Select order %d', 'dyapress-ecommerce' ), $o->id ).'</label><input id="cb-select-'.$o->id.'" type="checkbox" name="orders[]" value="'.$o->id.'">',
	    ),
	    'id' => array(
		'class' => 'manage-column column-id column-primary',
		'content' => '<span class="dyapress_ecommerce_order_id">'.$o->id.'</span><button type="button" class="toggle-row"><span class="screen-reader-text">'.__( 'Display more', 'dyapress-ecommerce' ).'</span></button>',
	    ),
	    'total' => dyapress_ecommerce()->format_price( $o->total_vat_incl, $o ),
	    'capturable' => $o->total_capturable > 0 ? dyapress_ecommerce()->format_price( $o->total_capturable, $o ) : '',
	    'paid' => $o->total_paid > 0 ? dyapress_ecommerce()->format_price( $o->total_paid, $o ) : '',
	    'status' => dyapress_ecommerce()->format_order_state_mark( $o->order_state ),
	    'date_created' => date_i18n( $date_format, $o->order_creation_date ),
	);

	if ( isset( $o->order_payment_type_ids ) ) {
	    foreach ( $payment_types as $id => $name ) {
		if ( 1 << $id & $o->order_payment_type_ids )
		    $p['payment'] .= ( $p['payment'] == '' ? '' : ', ' ).$name;
	    }
	} elseif ( isset( $o->order_payment_type_id ) && $o->order_payment_type_id > 0 ) {
	    $p['payment'] = $payment_types[$o->order_payment_type_id];
	}

	if ( $o->total_paid >= $o->total_vat_incl ) {
	    $p['price'] = '<mark class="paid">'.dyapress_ecommerce()->format_price( $o->total_vat_incl, $o ).'</mark>';
	} else {
	    $p['price'] = '<mark class="unpaid">'.dyapress_ecommerce()->format_price( $o->total_vat_incl, $o ).'</mark>';
	}

	if ( $o->user_id > 0 ) {
	    $user = dyapress_premium()->get_user( $o->user_id, false );

	    $p['user'] = array(
		'class' => 'manage-column column-user column-username',
		'content' => get_avatar( $o->user_id, 32 ) . '&nbsp;' . $user->firstname . ' ' . $user->lastname,
	    );

	    if ( $o->thirdparty_id > 0 && isset( $user->thirdparties[$o->thirdparty_id] ) ) {
		$thirdparty = $user->thirdparties[$o->thirdparty_id];

		if ( !empty( $thirdparty->company ) )
		    $p['thirdparty'] = $thirdparty->company;
		elseif ( !empty( $thirdparty->firstname ) && !empty( $thirdparty->lastname ) )
		    $p['thirdparty'] = $thirdparty->firstname.' '.$thirdparty->lastname;
		else
		    $p['thirdparty'] = "#".$o->thirdparty_id;
	    } elseif ( $o->thirdparty_id > 0 )
		$p['thirdparty'] = "#".$o->thirdparty_id;
	}

	if ( $o->order_reference && $o->order_id > 0 ) {
	    $l = dyapress_ecommerce()->document_order_permalink( $o );
	    if ($l)
		$p['order'] = '<a target="_blank" href="'.$l.'">'.$o->order_reference.'</a>';
	}

	if ( $o->order_invoice_reference && $o->order_invoice_id > 0 ) {
	    $l = dyapress_ecommerce()->document_order_invoice_permalink( $o );
	    if ($l)
		$p['invoice'] = '<a target="_blank" href="'.$l.'">'.$o->order_invoice_reference.'</a>';
	}

	return apply_filters( 'dyapress_ecommerce_admin_order_display', $p );
    }

    public function order_columns() {
	return apply_filters( 'dyapress_ecommerce_order_columns', array(
	    'id' => __( 'ID', 'dyapress-ecommerce' ),
	    'total' => __( 'Total', 'dyapress-ecommerce' ),
	    'paid' => __( 'Paid', 'dyapress-ecommerce' ),
	    'status' => __( 'Status', 'dyapress-ecommerce' ),
	    'date_created' => array(
		'class' => 'column-date column-date_created',
		'content' => __( 'Date', 'dyapress-ecommerce' ),
	    ),
	    'order' => __( 'Order', 'dyapress-ecommerce' ),
	    'invoice' => __( 'Invoice', 'dyapress-ecommerce' ),
	) );
    }

    public function order_display( $o, $date_format = null ) {
	if ( empty( $date_format ) )
	    $date_format = get_option( 'date_format' );

	$p = array(
	    'id' => '#<span class="dyapress_ecommerce_order_id">'.$o->id.'</span>',
	    'total' => dyapress_ecommerce()->format_price( $o->total_vat_incl, $o ),
	    'paid' => $o->total_paid > 0 ? dyapress_ecommerce()->format_price( $o->total_paid, $o ) : '',
	    'status' => dyapress_ecommerce()->format_order_state_mark( $o->order_state ),
	    'date_created' => date_i18n( $date_format, $o->order_creation_date ),
	);

	if ( $o->total_paid >= $o->total_vat_incl ) {
	    $p['price'] = '<mark class="paid">'.dyapress_ecommerce()->format_price( $o->total_vat_incl, $o ).'</mark>';
	} else {
	    $p['price'] = '<mark class="unpaid">'.dyapress_ecommerce()->format_price( $o->total_vat_incl, $o ).'</mark>';
	}

	if ( $o->user_id > 0 ) {
	    $user = dyapress_premium()->get_user( $o->user_id );

	    if ( $o->thirdparty_id > 0 && isset( $user->thirdparties[$o->thirdparty_id] ) ) {
		$thirdparty = $user->thirdparties[$o->thirdparty_id];

		if ( !empty( $thirdparty->company ) )
		    $p['thirdparty'] = $thirdparty->company;
		elseif ( !empty( $thirdparty->firstname ) && !empty( $thirdparty->lastname ) )
		    $p['thirdparty'] = $thirdparty->firstname.' '.$thirdparty->lastname;
	    } elseif ( $o->thirdparty_id > 0 )
		$p['thirdparty'] = "#".$o->thirdparty_id;
	}

	if ( $o->order_reference && $o->order_id > 0 ) {
	    $l = dyapress_ecommerce()->document_order_permalink( $o );
	    if ($l)
		$p['order'] = '<a target="_blank" href="'.$l.'">'.$o->order_reference.'</a>';
	}

	if ( $o->order_invoice_reference && $o->order_invoice_id > 0 ) {
	    $l = dyapress_ecommerce()->document_order_invoice_permalink( $o );
	    if ($l)
		$p['invoice'] = '<a target="_blank" href="'.$l.'">'.$o->order_invoice_reference.'</a>';
	    else
		$p['invoice'] = $o->order_invoice_reference;
	}

	return apply_filters( 'dyapress_ecommerce_order_display', $p );
    }

    public function proposal_columns() {
	return apply_filters( 'dyapress_ecommerce_proposal_columns', array(
	    'proposal' => __( 'Proposal', 'dyapress-ecommerce' ),
	    'total' => __( 'Total', 'dyapress-ecommerce' ),
	    'status' => __( 'Status', 'dyapress-ecommerce' ),
	    'creation_date' => array(
		'class' => 'column-date column-creation_date',
		'content' => __( 'Date', 'dyapress-ecommerce' ),
	    ),
	    'expiration_date' => array(
		'class' => 'column-date column-expiration_date',
		'content' => __( 'Expiration date', 'dyapress-ecommerce' ),
	    ),
	) );
    }

    public function proposal_display( $p, $date_format = null ) {
	if ( empty( $date_format ) )
	    $date_format = get_option( 'date_format' );

	$d = array(
	    'total' => dyapress_ecommerce()->format_price( $p->total_vat_incl, $p ),
	    'status' => dyapress_ecommerce()->format_proposal_status_mark( $p->status ),
	    'creation_date' => date_i18n( $date_format, $p->creation_date ),
	    'expiration_date' => date_i18n( $date_format, $p->expiration_date ),
	);

	if ( !empty( $p->reference ) ) {
	    $l = dyapress_ecommerce()->document_proposal_permalink( $p->id );
	    if ($l)
		$d['proposal'] = '<a target="_blank" href="'.$l.'">'.$p->reference.'</a>';
	    else
		$d['proposal'] = $p->reference;
	}

	return apply_filters( 'dyapress_ecommerce_proposal_display', $d );
    }

    public function invoice_columns() {
	return apply_filters( 'dyapress_ecommerce_invoice_columns', array(
	    'invoice' => __( 'Invoice', 'dyapress-ecommerce' ),
	    'total' => __( 'Total', 'dyapress-ecommerce' ),
	    'paid' => __( 'Paid', 'dyapress-ecommerce' ),
	    'date_created' => array(
		'class' => 'column-date column-date_created',
		'content' => __( 'Date', 'dyapress-ecommerce' ),
	    ),
	) );
    }

    public function invoice_display( $i, $date_format = null ) {
	if ( empty( $date_format ) )
	    $date_format = get_option( 'date_format' );

	$p = array(
	    'total' => dyapress_ecommerce()->format_price( $i->total_vat_incl, $i ),
	    'paid' => $i->total_paid > 0 ? dyapress_ecommerce()->format_price( $i->total_paid, $i ) : '',
	    'date_created' => date_i18n( $date_format, $i->creation_date ),
	);

	if ( !empty( $i->reference ) ) {
	    $l = dyapress_ecommerce()->document_invoice_permalink( $i->id );
	    if ($l)
		$p['invoice'] = '<a target="_blank" href="'.$l.'">'.$i->reference.'</a>';
	    else
		$p['invoice'] = $i->reference;
	}

	return apply_filters( 'dyapress_ecommerce_invoice_display', $p );
    }

    public function echo_columns_header( $columns, $tr = '' ) {
	foreach ( $columns as $c => $data ) {
	    if ( is_array( $data ) ) {
		$tag = !empty( $data['tag'] ) ? sanitize_key( $data['tag'] ) : 'th';
		$class = !empty( $data['class'] ) ? $data['class'] : 'column-'.sanitize_key( $c );

		echo '<'.$tag.' class="'.$class.'">'. $data['content'] . '</'.$tag.'>';
	    } else {
		echo '<th class="column-'.sanitize_key( $c ).'">'.$data.'</th>';
	    }
	}
    }

    public function echo_columns_body( $p, $columns, $tr = '' ) {
	if ( is_array( $p ) ) {
	    echo '<tr'.( !empty( $tr ) ? ' '.$tr : '' ).'>';
	    foreach ( $columns as $c => $data ) {
		$colname = is_array( $data ) ? $data['content'] : $data;

		if ( isset( $p[$c] ) && is_array( $p[$c] ) ) {
		    $tag = !empty( $p[$c]['tag'] ) ? sanitize_key( $p[$c]['tag'] ) : 'td';
		    $class = !empty( $p[$c]['class'] ) ? $p[$c]['class'] : 'column-'.sanitize_key( $c );
		    $content = !empty( $p[$c]['content'] ) ? $p[$c]['content'] : '';

		    echo '<'.$tag.( $tag == 'th' ? ' scope="row"' : '' ).' class="'.sanitize_text_field( $class ).'" data-colname="'.sanitize_text_field( $colname ).'">'. $content . '</'.$tag.'>';
		} else {
		    echo '<td class="column-'.sanitize_key( $c ).'" data-colname="'.sanitize_text_field( $colname ).'">'.( isset( $p[$c] ) ? $p[$c] : '' ).'</td>';
		}
	    }
	    echo '</tr>';
	}
    }

    public function echo_thirdparty( $t, $tag = 'h3' ) {
	$linefields = apply_filters( 'dyapress_ecommerce_thirdparty', array(
	    'company|firstname lastname',
	    'vat_number',
	    'firstname lastname',
	    'address',
	    'postcode city',
	    'country',
	) );

	$n = 0;
	foreach ( $linefields as $line ) {
	    $row = '';
	    $choices = explode( '|', $line );

	    foreach ( $choices as $c ) {
		$fields = explode( ' ', $c );

		foreach ( $fields as $f ) {
		    $k = sanitize_key( $f );

		    if ( dyapress_premium()->has_user_field( $k ) && !empty( $t->$k ) ) {
			    $row .= $t->$k.' ';
		    } else {
			if ( $k == 'country' && dyapress_premium()->has_user_field( 'country_id' ) ) {
			    if ( !empty( $t->country_id ) ) {
				$countries = dyapress_premium()->get_countries();

				foreach ( $countries as $country )
				    if ( $country->id == $t->country_id )
					$row .= $country->name;
			    }
			}
		    }
		}

		$row = trim( $row );
		if ( $row != '' ) {
		    if ( !$n )
			echo '<'.$tag.'>'.esc_html( $row ).'</'.$tag.'><address>';
		    elseif ( $n > 1 )
			echo '<br><span>'.esc_html( $row ).'</span>';
		    else
			echo '<span>'.esc_html( $row ).'</span>';

		    $n++;
		    break;
		} 
	    }
	}

	if ( $n > 0 )
	    echo '</address>';
    }

    public function dropdown_thirdparty( $user, $intro = '' ) {
	if ( is_array( $user ) )
	    $user = (object)$user;

	if ( !empty( $user ) && is_object( $user ) && !empty( $user->thirdparties ) ) {
	    echo $intro.'<span class="field"><select name="dyapress-ecommerce_thirdparty_id" class="thirdparty">';
	    foreach ( $user->thirdparties as $t ) {
		if ( !empty( $t->country_id ) ) {
		    echo '<option value="'.$t->thirdparty_id.'">';

		    if ( !empty( $t->company ) )
			echo esc_html( $t->company );
		    else
			echo esc_html( $t->firstname.' '.$t->lastname );

		    echo esc_html( ' - '.join( ', ', array_filter( array_map( 'trim', explode( "\n", $t->address ) ),
			function( $s ) { return !empty( $s ); } ) ).' '.$t->postcode.' '.$t->city );

		    if ( empty( $t->address ) && empty( $t->postcode ) && empty( $t->city ) ) {
			$countries = dyapress_premium()->get_countries();

			foreach ( $countries as $country )
			    if ( $country->id == $t->country_id )
				echo $country->name;
		    }
		}

		echo '</option>';
	    }
	    echo '<option value="0">'.__( 'Add new thirdparty', 'dyapress-ecommerce' ).'</option></select></span>';
	}
    }

    function document_order_permalink( $order ) {
	if ( ( $id = dyapress_premium()->get_order_id( $order ) ) ) {
	    return get_home_url() . '?dyapress_download_order='.$id.'&locale='.get_locale();
	}

	return false;
    }

    function document_order_invoice_permalink( $order ) {
	if ( ( $id = dyapress_premium()->get_order_id( $order ) ) ) {
	    return get_home_url() . '?dyapress_download_order_invoice='.$id.'&locale='.get_locale();
	}

	return false;
    }

    function document_proposal_permalink( $id ) {
	if ( isset( $id ) && (int)$id > 0 )
	    return get_home_url() . '?dyapress_download_proposal='.( (int)$id ).'&locale='.get_locale();
	else
	    return false;
    }

    function document_invoice_permalink( $id ) {
	if ( isset( $id ) && (int)$id > 0 )
	    return get_home_url() . '?dyapress_download_invoice='.( (int)$id ).'&locale='.get_locale();
	else
	    return false;
    }

    function register_order_state( $status, $label, $bg_color, $fg_color ) {
	if ( isset( $this->order_state[$status] ) )
	    $prev = $this->order_state[$status];
	else
	    $prev = null;

	$this->order_state[$status] = array(
	    'label' => $label,
	    'bg_color' => $bg_color,
	    'fg_color' => $fg_color,
	);

	return $prev;
    }

    function deregister_order_state( $status ) {
	if ( isset( $this->order_state[$status] ) ) {
	    $prev = $this->order_state[$status];

	    unset( $this->order_state[$status] );

	    return $prev;
	} else
	    return false;
    }

    function format_order_state( $status ) {
	if ( isset( $this->order_state[$status] ) ) {
	    $p = $this->order_state[$status];

	    return $p['label'];
	} else {
	    return $status;
	}
    }

    function format_order_state_mark( $status ) {
	if ( isset( $this->order_state[$status] ) ) {
	    $p = $this->order_state[$status];

	    return '<mark class="status" style="background-color: '.$p['bg_color'].'; color: '.$p['fg_color'].'">'.$p['label'].'</mark>';
	} else {
	    return $status;
	}
    }

    function format_proposal_status_mark( $status ) {
	$status_label = array(
	    'draft' => __( 'Draft', 'dyapress-ecommerce' ),
	    'validated' => __( 'Validated', 'dyapress-ecommerce' ),
	    'signed' => __( 'Signed', 'dyapress-ecommerce' ),
	    'closed' => __( 'Closed', 'dyapress-ecommerce' ),
	    'invoiced' => __( 'Invoiced', 'dyapress-ecommerce' ),
	);
	$status_bgcolor = array(
	    'draft' => '#e5e5e5',
	    'validated' => '#ddeeff',
	    'signed' => '#ffa020',
	    'closed' => '#f01040',
	    'invoiced' => '#108010',
	);
	if ( isset( $status_label[$status] ) ) {
	    $color = $status == 'draft' ? '#777' : 'white';

	    return '<mark class="status" style="background-color: '.$status_bgcolor[$status].'; color: '.$color.'">'.$status_label[$status].'</mark>';
	} else {
	    return $status;
	}
    }

    public function format_price( $price, $currency = null ) {
	if ( !empty( $price ) ) {
	    if ( is_object( $currency ) && !empty( $currency->currency ) )
		$currency = $currency->currency;
	    elseif ( empty( $currency ) )
		$currency = dyapress_premium()->default_currency();

	    return apply_filters( 'dyapress_ecommerce_format_price', '<span class="format-price price" data-price="'.esc_js( $price ).'" data-currency="'.$currency.'"></span>', $price, $currency );
		// <script>document.scripts[document.scripts.length-1].parentNode.innerHTML = (new Intl.NumberFormat((new Intl.NumberFormat()).resolvedOptions().locale,{style:"currency",currency:"'.esc_js( $currency ).'"}).format('.esc_js( $price ).'));</script></span>', $price, $currency );
	} else
	    return apply_filters( 'dyapress_ecommerce_format_price', '', $price, $currency );
    }

    public function format_price_vat_excl( $price, $currency = null, $sep = '&nbsp;' ) {
	// FIXME: VAT Excl position left or right ?
	return $this->format_price( $price, $currency ) . $this->format_vat_excl( $sep );
    }

    public function format_price_vat_incl( $price, $currency = null, $sep = '&nbsp;' ) {
	// FIXME: VAT Excl position left or right ?
	return $this->format_price( $price, $currency, $sep ) . $this->format_vat_incl( $sep );
    }

    public function format_vat_excl( $prefix = '&nbsp;', $suffix = '' ) {
	return apply_filters( 'dyapress_ecommerce_format_vat_excl', '<span class="use-vat">' . $prefix. __( '(VAT excluded)', 'dyapress-ecommerce' ) . $suffix . '</span>' );
    }

    public function format_vat_incl( $prefix = '&nbsp;', $suffix = '' ) {
	return apply_filters( 'dyapress_ecommerce_format_vat_incl', '<span class="use-vat">' . $prefix. __( '(VAT included)', 'dyapress-ecommerce' ) . $suffix . '</span>' );
    }

    /**
     * Init administration scripts for WordPress admin.
     */
    public function admin_enqueue_scripts() {
    	wp_enqueue_style( 'dyapress-ecommerce-admin-style', plugins_url( 'dyapress-ecommerce/css/admin.css' ), false, DYAPRESS_ECOMMERCE_VERSION );
    }

    function enqueue_scripts() {
    	wp_enqueue_style( 'dyapress-ecommerce-style', plugins_url( 'dyapress-ecommerce/css/style.css' ), false, DYAPRESS_ECOMMERCE_VERSION );

    	wp_enqueue_script( 'jquery' );
    	wp_enqueue_script( 'jquery-ui-tabs' );

	if ( dyapress_premium()->is_stripe_available() )
	    wp_enqueue_script( 'stripe-js', 'https://js.stripe.com/v3/', array(), null);
    }

    function activate() {
	$this->options = get_option( 'dyapress_ecommerce_options' );

	if ($this->options == false) {
	    $this->options = array();

	    add_option( 'dyapress_ecommerce_options', $this->options );
	} else {
	    $this->options = $this->validate_options( $this->options );

	    update_option( 'dyapress_ecommerce_options', $this->options );
	}

	remove_role( 'customer' );
	add_role(
	    'customer',
	    __( 'Customer', 'dyapress-ecommerce' ),
	    array(
		'read' => true,
	    )
	);

	$installed_db_version = get_option( 'dyapress_ecommerce_db_version' );
	// error_log("installed_db_version=$installed_db_version, DYAPRESS_ECOMMERCE_VERSION=".DYAPRESS_ECOMMERCE_VERSION);
	if ( $installed_db_version != DYAPRESS_ECOMMERCE_VERSION ) {
	    global $wpdb;

	    $collate = '';
	    if ( $wpdb->has_cap( 'collation' ) )
		$collate = $wpdb->get_charset_collate();

	    $tableprefix = $wpdb->prefix . 'dyapress_ecommerce_';

	    $maillogtable = $tableprefix . 'maillog';
	    $sql = "CREATE TABLE " . $maillogtable . " (
		`id` BIGINT(20) NOT NULL AUTO_INCREMENT,
		`date` DATETIME NOT NULL,
		`slug` VARCHAR(50) NOT NULL,
		`to` VARCHAR(255) NOT NULL,
		`subject` VARCHAR(255) NOT NULL,
		`body` LONGTEXT,
		PRIMARY KEY  (`id`)
	    ) " . $collate . ";";

	    require_once( ABSPATH . 'wp-admin/includes/upgrade.php' );

	    dbDelta($sql);

	    update_option( 'dyapress_ecommerce_db_version', DYAPRESS_PREMIUM_VERSION );
	}
    }

    function deactivate() {
	remove_role( 'customer' );

	//delete_option( 'dyapress_ecommerce_options' );
	delete_option( 'dyapress_ecommerce_db_version' );
    }

    function register_options() {
	register_setting( 'dyapress_ecommerce_group', 'dyapress_ecommerce_options', array( &$this, 'validate_options' ) );
    }

    function validate_options( $input ) {

	//$input[ 'licence_key' ] = $input[ 'licence_key' ]; #TODO licence ?

	return $input;
    }

    public function add_account_section( $section_title, $name, $content, $position = null ) {
        $new_section = array(
                'title' => $section_title,
                'name' => $name,
                'content' => $content,
                );
        if ( null === $position ) {
            $this->account_sections[] = $new_section;
	} elseif ( isset( $this->account_sections[ "$position" ] ) ) {
	 	$position = $position + substr( base_convert( md5( $new_section['title'] ), 16, 10 ) , -5 ) * 0.00001;
		$this->account_sections[ "$position" ] = $new_section;
	} else {
            $this->account_sections[ $position ] = $new_section;
	}

        return $new_section;
    }

    public function get_account_sections() {
        if ( !did_action( 'dyapress_ecommerce_add_account_sections' ) ) {
	    require_once DYAPRESS_ECOMMERCE_PLUGIN_PATH . '/includes/class-myaccount.php';

            do_action( 'dyapress_ecommerce_add_account_sections', dyapress_premium()->get_current_user() );
        }

        ksort( $this->account_sections, SORT_NUMERIC );

        return apply_filters( 'dyapress_ecommerce_filter_account_sections', $this->account_sections );
    }

    // Decode args which is an associative array of 
    // * products to get the list of product_id separated by a ";" that can be accepted by dyapress_premium()->get_product()
    // * products_quantity formatted as a list of the following separated by a ";" :
    //   - product_id => quantity of product, must be 0 or positive, default is 1
    //   - product_id=quantity => quantity of product is fixed to quantity
    //   - product_id=min-max => quantity of product is between min and max, default is min or 1
    // * products_label formatted as a list of the following separated by ";" :
    //   - product_id => label of product, if not defined the label will be the name of the product
    // * products_post formatted as a list of the following separated by ";" :
    //   - product_id=post_id => post_id is the ID or the post_name of a post
    // * products_hidden formatted as a list of the following separated by ";" :
    //   - product_id => the product will not be visible in a form
    //
    // order of the result products will be given in the same order of options and list of product_id found
    public function decode_products( $args, $caching = true ) {
	$result = array();

	if ( isset( $args['products'] ) ) {
	    $products = is_array( $args['products'] ) ? $args['products'] : explode( ';', $args['products'] );

	    foreach ( $products as $pid ) {
		if ( $p = dyapress_premium()->get_product( $pid, $caching ) ) {
		    $result[$p->id] = array(
			'pid' => $pid,
			'product' => $p,
		    );
		} else
		    $result[$pid] = null;
	    }
	}

	if ( isset( $args['products_quantity'] ) ) {
	    $products_quantity = is_array( $args['products_quantity'] ) ? $args['products_quantity'] : explode( ';', $args['products_quantity'] );

	    foreach ( $products_quantity as $pid ) {
		if ( preg_match( '/^\s*([^=]*)=\s*(\d*)\s*-\s*(\d*)\s*$/', $pid, $matches ) ) {
		    if ( $p = dyapress_premium()->get_product( $matches[1], $caching ) ) {
			if ( isset( $result[$p->id] ) ) {
			    $result[$p->id]['min'] = !empty( $matches[2] ) ? intval( $matches[2] ) : 0;
			    $result[$p->id]['max'] = !empty( $matches[3] ) ? intval( $matches[3] ) : 0;
			    $result[$p->id]['quantity'] = !empty( $matches[2] ) ? intval( $matches[2] ) : 1;
			} else {
			    $result[$p->id] = array(
				'pid' => $matches[1],
				'product' => $p,
				'min' => !empty( $matches[2] ) ? intval( $matches[2] ) : 0,
				'max' => !empty( $matches[3] ) ? intval( $matches[3] ) : 0,
				'quantity' => !empty( $matches[2] ) ? intval( $matches[2] ) : 1,
			    );
			}
		    } else
			$result[$matches[1]] = null;
		} elseif ( preg_match( '/^\s*([^=]*)=\s*(\d+)\s*$/', $pid, $matches ) ) {
		    if ( $p = dyapress_premium()->get_product( $matches[1], $caching ) ) {
			if ( isset( $result[$p->id] ) ) {
			    $result[$p->id]['min'] = !empty( $matches[2] ) ? intval( $matches[2] ) : 0;
			    $result[$p->id]['max'] = !empty( $matches[2] ) ? intval( $matches[2] ) : 0;
			    $result[$p->id]['quantity'] = !empty( $matches[2] ) ? intval( $matches[2] ) : 1;
			} else {
			    $result[$p->id] = array(
				'pid' => $matches[1],
				'product' => $p,
				'min' => !empty( $matches[2] ) ? intval( $matches[2] ) : 0,
				'max' => !empty( $matches[2] ) ? intval( $matches[2] ) : 0,
				'quantity' => !empty( $matches[2] ) ? intval( $matches[2] ) : 1,
			    );
			}
		    } else
			$result[$matches[1]] = null;
		} else {
		    if ( $p = dyapress_premium()->get_product( $pid, $caching ) ) {
			if ( isset( $result[$p->id] ) ) {
			} else {
			    $result[$p->id] = array(
				'pid' => $pid,
				'product' => $p,
			    );
			}
		    } else
			$result[$matches[1]] = null;
		}
	    }
	}

	if ( isset( $args['products_label'] ) ) {
	    $products_label = is_array( $args['products_label'] ) ? $args['products_label'] : explode( ';', $args['products_label'] );

	    foreach ( $products_label as $pid ) {
		if ( preg_match( '/^\s*([^=]*)=\s*(.*+)\s*$/', $pid, $matches ) ) {
		    if ( $p = dyapress_premium()->get_product( $matches[1], $caching ) ) {
			if ( isset( $result[$p->id] ) ) {
			    $result[$p->id]['label'] = $matches[2];
			} else {
			    $result[$p->id] = array(
				'pid' => $matches[1],
				'product' => $p,
				'label' => $matches[2],
			    );
			}
		    } else
			$result[$matches[1]] = null;
		} else
		    $result[$pid] = null;
	    }
	}

	if ( isset( $args['products_post'] ) ) {
	    $products_post = is_array( $args['products_post'] ) ? $args['products_post'] : explode( ';', $args['products_post'] );

	    foreach ( $products_post as $pid ) {
		if ( preg_match( '/^\s*([^=]*)=\s*(?:(\w+):)?(\S+)\s*$/', $pid, $matches ) ) {
		    if ( $p = dyapress_premium()->get_product( $matches[1], $caching ) ) {
			$args = array(
			    'numberposts' => 1,
			    'post_status' => 'publish',
			);
			if ( !empty( $matches[2] ) )
			    $args['post_type'] = $matches[2];
			if ( is_numeric( $matches[3] ) )
			    $args['p'] = $matches[3];
			else
			    $args['name'] = $matches[3];

			$query = new WP_Query( $args );
			$posts = $query->get_posts();

			if ( isset( $result[$p->id] ) ) {
			    $result[$p->id]['post'] = array_shift( $posts );
			} else {
			    $result[$p->id] = array(
				'pid' => $matches[1],
				'product' => $p,
				'post' => array_shift( $posts ),
			    );
			}
		    } else
			$result[$matches[1]] = null;
		} else
		    $result[$pid] = null;
	    }
	}

	if ( isset( $args['products_hidden'] ) ) {
	    $products_hidden = is_array( $args['products_hidden'] ) ? $args['products_hidden'] : explode( ';', $args['products_hidden'] );

	    foreach ( $products_label as $pid ) {
		if ( $p = dyapress_premium()->get_product( $pid, $caching ) ) {
		    if ( isset( $result[$p->id] ) ) {
			$result[$p->id]['hidden'] = true;
		    } else {
			$result[$p->id] = array(
			    'pid' => $matches[1],
			    'product' => $p,
			    'hidden' => true,
			);
		    }
		} else
		    $result[$pid] = null;
	    }
	}

	foreach ( $result as $pid => &$p ) {
	    // check for default values and convert to object at the end
	    if ( isset( $p ) ) {
		if ( !isset( $p['min'] ) )
		    $p['min'] = 0;
		if ( !isset( $p['max'] ) )
		    $p['max'] = 0;
		if ( !isset( $p['quantity'] ) )
		    $p['quantity'] = 1;
		if ( !isset( $p['label'] ) )
		    $p['label'] = $p['product']->label;

		$p = (object)$p;
	    }
	}

	return $result;
    }

    public function is_myaccount_page( $id = false ) {
	if ( !$id )
	    $id = get_the_ID();

	if ( empty( $this->options ) )
	    $this->options = get_option( 'dyapress_ecommerce_options' );

	if ( !empty( $this->options ) && isset( $this->options['myaccount_page_id'] ) ) {
	    $myaccount_page_id = $this->options['myaccount_page_id'];

	    if ( (int)$myaccount_page_id == (int)$id )
		return true;
	    else {
		global $polylang;

		if ( isset( $polylang ) && isset( $polylang->model ) ) {
		    if ( in_array( (int)$myaccount_page_id, $polylang->model->get_translations( 'post', (int)$id ), true ) ) {
			return true;
		    }
		}
	    }
	}

	return false;
    }

    public function display_post_states( $states, $post ) {
	if ( $this->is_myaccount_page( $post->ID ) )
	    $states['dyapress_ecommerce_myaccount'] = __( 'My account page', 'dyapress-ecommerce' );

	return $states;
    }

    public function myaccount_content( $content ) {
	if ( in_the_loop() )
	    return $this->is_myaccount_page() && preg_match( '/^\s*$/', $content ) ? DyaPressECommerceShortcodes::myaccount( array() ) : $content;
	else
	    return $content;
    }

    public function get_myaccount_permalink() {
	if ( empty( $this->options ) )
	    $this->options = get_option( 'dyapress_ecommerce_options' );

	$myaccount_page_id = $this->options['myaccount_page_id'];
	if ( function_exists( 'pll_get_post' ) )
	    $myaccount_page_id = pll_get_post( $myaccount_page_id );

	return get_permalink( $myaccount_page_id );
    }

    public function get_myaccount_activation_permalink( $param, $args = null, $token_name = 'activate' ) {
	if ( !empty( $param ) ) {
	    $permalink = $this->get_myaccount_permalink();
	    if ( empty( $permalink ) )
		return null;

	    if ( is_object( $param ) )
		$param = (array)$param;

	    $param = dyapress_premium()->validate_user( $param );

	    // email is the only mandatory field to allow activation verification
	    if ( !empty( $param ) && !empty( $param->email ) ) {
		if ( !empty( $args ) ) {
		    foreach ( $args as $k => $v )
			$param->$k = $v;
		}

		$param->nonce = wp_create_nonce( 'activate_account' );

		return $permalink.'?'.$token_name.'='.urlencode( base64_encode( http_build_query( $param ) ) );
	    }
	}

	return null;
    }

    public function verify_myaccount_activation_token( $token ) {
	parse_str( base64_decode( $token ), $data );

	return isset( $data['nonce'] ) && wp_verify_nonce( $data['nonce'], 'activate_account' ) ? $data : null;
    }

    public function send_mail( $to, $subject, $mailslug, $args = array(), $attachments = array() ) {
	if ( isset( $to ) && isset( $subject ) && isset( $mailslug ) ) {
	    if ( is_object( $to ) )
		$to = $to->email;

	    $headers = array(
		'Content-Type: text/html; charset='.get_option( 'blog_charset' ),
	    );

	    if ( !empty( $this->options['mail_from_email'] ) && filter_var( $this->options['mail_from_email'], FILTER_VALIDATE_EMAIL ) ) {
		if ( !empty( $this->options['mail_from_name'] ) )
		    $from = '"'.$this->options['mail_from_name'].'" ';
		else
		    $from = '';

		$headers[] = 'From: '.$from.'<'.$this->options['mail_from_email'].'>';
	    }

	    $body = $this->get_template_part( 'mails/'.$mailslug, null, $args );

	    if ( wp_mail( $to, $subject, $body, $headers, $attachments ) ) {

		global $wpdb;

		return $wpdb->insert( $wpdb->prefix . 'dyapress_ecommerce_maillog',
		    array (
			'date' => date( 'Y-m-d H:i:s' ),
			'slug' => $mailslug,
			'to' => $to,
			'subject' => $subject,
			'body' => $body,
		    ), array( '%s', '%s', '%s', '%s', '%s' ) );
	    } else
		return false;
	}

	return null;
    }

    public function filter_create_order_payment( $meta ) {
	return array_merge( $meta, array(
	    'stripe_payment_type_id' => $this->stripe_payment_type_id(),
	    'stripe_payment_account_id' => $this->stripe_payment_account_id(),
	) );
    }

    public function template_path() {
	return apply_filters( 'dyapress_ecommerce_template_path', 'dyapress/' );
    }

    public function locate_template_part( $slug, $name = null ) {
	if ( !empty( $slug ) ) {
	    $templates = array();

	    if ( isset( $name ) )
		$templates[] = dyapress_ecommerce()->template_path() . $slug . '-' . $name . '.php';
	    $templates[] = dyapress_ecommerce()->template_path() . $slug . '.php';

	    $filename = locate_template( $templates );

	    if ( !$filename )
		$filename = DYAPRESS_ECOMMERCE_PLUGIN_PATH . '/templates/' . $slug . '.php';

	    if ( file_exists( $filename ) )
		return $filename;
	}

	return null;
    }

    public function template_part( $slug, $name = null, $args = null ) {
	if ( $filename = $this->locate_template_part( $slug, $name ) ) {
	    do_action( 'dyapress_ecommerce_before_template_part', $slug, $name, $args );

	    if ( isset( $slug ) ) { global $dyapress_ecommerce_slug; $dyapress_ecommerce_slug = $slug; }
	    if ( isset( $name ) ) { global $dyapress_ecommerce_name; $dyapress_ecommerce_name = $name; }
	    if ( isset( $args ) ) { global $dyapress_ecommerce_args; $dyapress_ecommerce_args = $args; }

	    global $dyapress_ecommerce_current_user;
	    if ( !isset( $dyapress_ecommerce_current_user ) )
		$dyapress_ecommerce_current_user = dyapress_premium()->get_current_user();

	    include $filename;

	    do_action( 'dyapress_ecommerce_after_template_part', $slug, $name, $args );
	}
    }

    public function get_template_part( $slug, $name = null, $args = null ) {
	ob_start();

	$this->template_part( $slug, $name, $args );

	return ob_get_clean();
    }

    function options( $args = null ) {
	if ( empty( $this->options ) )
	    $this->options = get_option( 'dyapress_ecommerce_options' );

	if ( !empty( $args ) ) {
	    if ( is_array( $args ) ) {
		if ( isset( $args['stripe_payment_type_id'] ) )
		    $this->options['stripe_payment_type_id'][dyapress_premium()->uuid()] = (int)$args['stripe_payment_type_id'];
		if ( isset( $args['stripe_payment_account_id'] ) )
		    $this->options['stripe_payment_account_id'][dyapress_premium()->uuid()] = (int)$args['stripe_payment_account_id'];

		if ( isset( $args['myaccount_page_id'] ) )
		    $this->options['myaccount_page_id'] = (int)$args['myaccount_page_id'];
		if ( isset( $args['display_account_proposals'] ) )
		    $this->options['display_account_proposals'] = (int)$args['display_account_proposals'];
		if ( isset( $args['display_account_invoices'] ) )
		    $this->options['display_account_invoices'] = (int)$args['display_account_invoices'];
		if ( isset( $args['mail_from_name'] ) )
		    $this->options['mail_from_name'] = sanitize_text_field( $args['mail_from_name'] );
		if ( isset( $args['mail_from_email'] ) && filter_var( $args['mail_from_email'], FILTER_VALIDATE_EMAIL ) )
		    $this->options['mail_from_email'] = $args['mail_from_email'];

		update_option( 'dyapress_ecommerce_options', $this->options );
	    } else
		return null;
	}

	return $this->options;
    }

    function stripe_payment_type_id() {
	if ( empty( $this->options ) )
	    $this->options();

	return isset( $this->options['stripe_payment_type_id'] ) ? (int)$this->options['stripe_payment_type_id'][dyapress_premium()->uuid()] : null;
    }

    function stripe_payment_account_id() {
	if ( empty( $this->options ) )
	    $this->options();

	return isset( $this->options['stripe_payment_account_id'] ) ? (int)$this->options['stripe_payment_account_id'][dyapress_premium()->uuid()] : null;
    }

    function display_account_proposals_section() {
	if ( empty( $this->options ) )
	    $this->options();

	return $this->options['display_account_proposals'];
    }

    function display_account_invoices_section() {
	if ( empty( $this->options ) )
	    $this->options();

	return $this->options['display_account_invoices'];
    }
}

add_action( 'plugins_loaded', array( 'DyaPressECommerce', 'get_instance' ) );
