<?php
/**
 * DyaPress Premium.
 *
 * @package DyaPressPremium
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * WordPress DyaPress Premium library :
 * Copyright (C) 2018-2020      François Pons        <fpons@aytechnet.fr>
 *
 * Please see https://dyapress.com/
 */

defined( 'ABSPATH' ) or die( 'Oops' );

define( 'DYAPRESS_PREMIUM_VERSION', '0.7' );

/**
 * Main DyaPress Premium Class.
 *
 * @class DyaPressPremium
 */
final class DyaPressPremium {
    private static $instance = null;

    private $options;
    private $country;
    private $european_countries = array();
    private $user_meta_fields = array();

    /**
     * Main DyaPress Premium API Instance.
     *
     * Ensures only one instance of DyaPress Premium is loaded or can be loaded.
     *
     * @static
     * @return DyaPressPremium - Main instance.
     */
    static function get_instance() {
	if ( is_null( self::$instance ) ) {
	    self::$instance = new DyaPressPremium();
	}

	return self::$instance;
    }

    /**
     * DyaPressPremiumECommerce Constructor.
     */
    private function __construct() {
	$this->options = get_option( 'dyapress_premium' );
	if ( empty( $this->options ) )
	    $this->options = array();

	$this->register_user_field( 'address', 'str' );
	$this->register_user_field( 'postcode', 'str' );
	$this->register_user_field( 'city', 'str' );
	$this->register_user_field( 'company', 'str', 'name' );
	$this->register_user_field( 'vat_number', 'str' );
	$this->register_user_field( 'country_id', 'int' );

	do_action( 'dyapress_premium_init' );
    }

    /**
     * Get API UUID of this instance.
     *
     * @api
     *
     * @return string The premium UUID of this instance.
     */
    public function premium_uuid() {
	if ( empty( $this->options['premium_uuid'] ) )
	    $this->options();

	return $this->options['premium_uuid'];
    }

    /**
     * Get API mode of this instance, which can be 'live' or 'test'.
     *
     * @api
     *
     * @return string The premium mode of this instance.
     */
    function mode() {
	if ( empty( $this->options['premium_uuid'] ) || empty( $this->options['expire'] ) || $this->options['expire'] < time() )
	    $this->options();

	return $this->options['mode'];
    }

    /**
     * Get UUID of used DyaPress or Dolibarr by this instance.
     *
     * @api
     *
     * @return string The UUID of used DyaPress or Dolibarr by this instance.
     */
    public function uuid() {
	if ( empty( $this->options['premium_uuid'] ) || empty( $this->options['expire'] ) || $this->options['expire'] < time() )
	    $this->options();

	return $this->options[$this->options['mode'].'_uuid'];
    }

    /**
     * Get Stripe application ID of DyaPress Premium according to current mode.
     *
     * @api
     *
     * @return string The Stripe application ID of DyaPress Premium.
     */
    public function stripe_application_id() {
	return 'ca_GMyVChLxnsJGx99CrYrxvmqoIHiLu1k8';
    }

    /**
     * Get Stripe account ID of the connected Stripe account with this instance.
     *
     * @api
     *
     * @return string|empty The Stripe account ID of the connected Stripe account with this instance, or an empty value if Stripe is not connected.
     */
    public function stripe_account_id() {
	if ( empty( $this->options['premium_uuid'] ) || empty( $this->options['expire'] ) || $this->options['expire'] < time() )
	    $this->options();

	return $this->options['stripe_account_id'];
    }

    /**
     * Get Stripe public key of the connected Stripe account with this instance.
     *
     * @api
     *
     * @return string|empty The Stripe public key of the connected Stripe account with this instance, or an empty value if Stripe is not connected.
     */
    public function stripe_public_key() {
	if ( empty( $this->options['premium_uuid'] ) || empty( $this->options['expire'] ) || $this->options['expire'] < time() )
	    $this->options();

	return $this->options['stripe_public_key'];
    }

    /**
     * Determines whether Stripe is available on front-end.
     *
     * @api
     *
     * @return bool True if Stripe is connected with this instance and SSL is active, false if not.
     */
    public function is_stripe_available() {
	if ( empty( $this->options['premium_uuid'] ) || empty( $this->options['expire'] ) || $this->options['expire'] < time() )
	    $this->options();

	return !empty( $this->options['stripe_public_key'] ) && is_ssl();
    }

    /**
     * Connect Stripe with this instance.
     *
     * @api
     *
     * @param string $code Code returned by Stripe after a successful OAuth connexion.
     *
     * @return object|null Result of DyaPress Premium connexion with Stripe, or null on error.
     */
    public function stripe_connect( $code ) {
	if ( !empty( $code ) ) {
	    if ( $this->is_active() )
		$result = $this->call_api('POST', '/v1/stripe/connect', array( 'code' => $code ) );

	    if ( isset( $result ) ) {
		// FIXME: update internal options with stripe_account_id and stripe_public_key
		$this->options['stripe_account_id'] = $result->stripe_account_id;
		$this->options['stripe_public_key'] = $result->stripe_public_key;

		update_option( 'dyapress_premium', apply_filters( 'dyapress_premium_options_update', $this->options ) );

		return $result;
	    }
	}

	return null;
    }

    /**
     * Compute Stripe state with a redirection URL and a custom state, generally a nonce.
     *
     * @api
     *
     * @param string $redirect_uri The landing page after returning from Stripe OAuth.
     * @param string $state State that will be added to $redirect_uri as 'state' GET variable.
     *
     * @return string Encoded string as state suitable for Stripe.
     */
    public function stripe_connect_state( $redirect_uri, $state ) {
	return urlencode( base64_encode( json_encode( array( 'redirect_uri' => $redirect_uri, 'state' => $state ) ) ) );
    }

    /**
     * Retrieve specific substring value according to UUID currently in use.
     *
     * This function extract from the encoded string the specific value associated to UUID or the default value.
     *
     * An example format of the encoded string for encoding 'val1' with 'UUID1', 'val2' with 'UUID2' and a default value 'val_default' is 'UUID1:val1,UUID:val2,val_default'.
     *
     * @api
     *
     * @param string $s The encoded string.
     *
     * @return string Specific value associated to UUID or the default value.
     */
    public function map_uuid( $s ) {
	if ( isset( $s ) ) {
	    if ( strpos( $s, ':' ) !== false ) {
		$uuid = $this->uuid();

		// search for a specific value for this uuid, or try to find the default value which must be a end of string
		if ( preg_match( '/\b'.$uuid.'\s*:\s*([^,]*)/', $s, $matches ) )
		    $s = $matches[1];
		elseif ( preg_match( '/,([^,:]*)$/', $s, $matches) )
		    $s = $matches[1];
		else
		    return null;
	    }
	}

	return $s;
    }

    /**
     * Encode value according to UUID currently in use.
     *
     * This function encodes a new encoded string according to the specific value associated to UUID currently in use.
     *
     * If a default value is needed, it must be set before using this function.
     *
     * @api
     *
     * @param string $s The encoded string.
     * @param string $v The value to be encoded with UUID currently in use.
     *
     * @return string The new encoded string.
     */
    public function set_map_uuid( $s, $v ) {
	$uuid = $this->uuid();

	if ( empty( $s ) )
	    return $uuid.':'.$v;

	if ( preg_match( '/\b'.$uuid.'\s*:\s*([^,]*)/', $s ) )
	    return preg_replace( '/\b'.$uuid.'\s*:\s*([^,]*)/', $v, $s );
	else
	    return $uuid.':'.$v.','.$s;
    }

    /**
     * Determines whether an external order notified is trusted originating from DyaPress Premium Microservice.
     *
     * @api
     *
     * @param object $order Order given by the notification.
     * @param string $premium_uuid API UUID of the notification.
     * @param string $secret A signature computed with API UUID, internal secret of this instance and order id.
     *
     * @return object|false|null Order is returned if trusted, false if order is untrusted and null on error.
     */
    public function sanitize_order( $order, $premium_uuid, $secret ) {
	// FIXME: sanitize_order() can only be called when options() has already been called at least once
	error_log('sanitize_order1');
	if ( empty( $this->options['premium_uuid'] ) || empty( $this->options['secret'] ) )
	    return null;

	error_log('sanitize_order2');
	if ( !empty( $order ) ) {
	    if ( !is_object( $order ) )
		$order = json_decode( $order );
	}

	error_log('sanitize_order3');
	if ( !empty( $order ) ) {
	    error_log("sanitize_order( premium_uuid=$premium_uuid, secret=$secret )");
	    if ( !empty( $premium_uuid ) && $this->options['premium_uuid'] == $premium_uuid &&
		 !empty( $secret ) && md5( $this->options['premium_uuid'].':'.$this->options['secret'].':'.$order->id ) == $secret )
		return $order;
	    else
		return false;
	} else
	    return null;
    }

    /**
     * Performs an API request to DyaPress Premium microservice.
     *
     * @api
     *
     * @param string $method HTTP method to use among GET, POST, PUT... 
     * @param string $url URL path to DyaPress Premium microservice including API version, like '/v1/orders'.
     *
     * @return mixed Result of API call already decoded or null on error.
     */
    public function call_api( $method, $url, $body_args = '' ) {
	if ( empty( $this->options['secret'] ) )
	    $this->options();

	$args = array(
	    'timeout' => 8,
	    'method' => $method,
	    'headers' => array(
		'Content-Type' => 'application/json; charset=utf8',
	    ),
	    'user-agent' => 'DyaPress-Premium/'.DYAPRESS_PREMIUM_VERSION.'; '.site_url(),
	);
	if ( !empty( $this->options['premium_uuid'] ) )
	    $args['headers']['DYAPRESS_PREMIUM_UUID'] = $this->options['premium_uuid'];
	if ( !empty( $this->options['secret'] ) )
	    $args['headers']['DYAPRESS_PREMIUM_SECRET'] = $this->options['secret'];

	if ( $body_args ) {
	    $args['body'] = json_encode($body_args);
	    $args['data_format'] = 'body';
	}

	$start_time = microtime(true);
	$response = wp_safe_remote_request( 'https://api.dyapress.com'.$url, $args );
	$end_time = microtime(true);

	if ( is_wp_error( $response ) ) {
	    error_log("DyaPress Premium API ($method $url) failed, execution time: ".($end_time - $start_time)." sec, error: ".$response->get_error_message());
	    return null;
	}

	$statuscode = wp_remote_retrieve_response_code( $response );
	if ( $statuscode == 200 || $statuscode == 202 ) {
	    if ( preg_match( '#application/json#', wp_remote_retrieve_header( $response, 'content-type' ) ) )
		$result = json_decode( wp_remote_retrieve_body( $response ) );
	    else
		$result = wp_remote_retrieve_body( $response );

	    error_log("DyaPress Premium API ($method $url) succeeded, execution time: ".($end_time - $start_time)." sec");
	} else {
	    $result = null;

	    error_log("DyaPress Premium API ($method $url) returned status=$statuscode, execution time: ".($end_time - $start_time)." sec");
	}

	return $result;
    }

    /**
     * Retrieves or set options of this instance.
     *
     * @api
     *
     * @param array $args Options to set.
     *
     * @return array Options updated.
     */
    function options( $args = null ) {
	// premium_uuid may be lost, because it is updated after a call to fetch premium licence information
	// but secret should not be lost as it used to verify premium api calls are allowed
	if ( empty( $this->options['secret'] ) ) {
	    // force update of secret both in DyaPress Premium area and here
	    $this->options['premium_uuid'] = null;
	    $this->options['secret'] = str_shuffle(md5('dyapress/'.DYAPRESS_PREMIUM_VERSION.'; '.site_url().'; '.microtime()));
	}

	$now = time();

	if ( empty( $this->options['premium_uuid'] ) || empty( $this->options['expire'] ) || $this->options['expire'] < $now || !empty( $args ) ) {
	    $data = array (
		'secret' => $this->options['secret'],
		'cb_url' => admin_url( 'admin-ajax.php' )
	    );

	    if ( !empty( $this->options['premium_uuid'] ) )
		$data['premium_uuid'] = $this->options['premium_uuid'];

	    if ( isset( $args ) && is_array( $args ) ) {
		if ( !empty( $args['test_uuid'] ) && $this->is_uuid( $args['test_uuid'] ) )
		    $data['test_uuid'] = $args['test_uuid'];
		if ( !empty( $args['test_dolapikey'] ) && $this->is_dolapikey( $args['test_dolapikey'] ) )
		    $data['test_dolapikey'] = $args['test_dolapikey'];
		if ( !empty( $args['live_uuid'] ) && $this->is_uuid( $args['live_uuid'] ) )
		    $data['live_uuid'] = $args['live_uuid'];
		if ( !empty( $args['live_dolapikey'] ) && $this->is_dolapikey( $args['live_dolapikey'] ) )
		    $data['live_dolapikey'] = $args['live_dolapikey'];
		if ( !empty( $args['country_code'] ) )
		    $data['country_code'] = strtoupper( sanitize_key( $args['country_code'] ) );
		if ( !empty( $args['mode'] ) ) {
		    if ( $args['mode'] == 'live' )
			$data['mode'] = 'live';
		    else
			$data['mode'] = 'test';
		}
	    }

	    $result = $this->call_api('POST', '/v1/premiumlicences', $data );

	    if ( isset( $result ) ) {
		// update expiration time of transient options
		if ( empty( $result->premium_expire_date ) ) {
		    $this->options['expire'] = $now + 4 * HOUR_IN_SECONDS;
		} else {
		    $this->options['expire'] = $result->premium_expire_date;
		    if ( $this->options['expire'] <= $now || $this->options['expire'] > $now + 4 * HOUR_IN_SECONDS )
			$this->options['expire'] = $now + 4 * HOUR_IN_SECONDS;
		}

		if ( !empty( $result->uuid ) )
		    $this->options['premium_uuid'] = $result->uuid;

		// fully rebuild options
		$this->options = array(
		    'premium_uuid' => $this->options['premium_uuid'],
		    'secret' => $this->options['secret'],
		    'expire' => $this->options['expire'],

		    'licence_level' => $result->licence_level,
		    'premium_expire_date' => $result->premium_expire_date,
		    'premium_available_api_calls' => $result->premium_available_api_calls,
		    'test_uuid' => $result->test_uuid,
		    'test_dolapikey' => $result->test_dolapikey,
		    'test_status' => $result->test_status,
		    'test_version' => $result->test_version,
		    'test_version_date' => $result->test_version_date,
		    'live_uuid' => $result->live_uuid,
		    'live_dolapikey' => $result->live_dolapikey,
		    'live_status' => $result->live_status,
		    'live_version' => $result->live_version,
		    'live_version_date' => $result->live_version_date,
		    'mode' => $result->mode,
		    'country_code' => $result->country_code,
		    'stripe_account_id' => $result->stripe_account_id,
		    'stripe_public_key' => $result->stripe_public_key,
		    'auth' => $result->auth,
		);

		update_option( 'dyapress_premium', apply_filters( 'dyapress_premium_options_update', $this->options ) );

		// make sure some transient are deleted on option changes
		delete_transient( 'dyapress_premium_payment_types' );
		delete_transient( 'dyapress_premium_bankaccounts' );
		delete_transient( 'dyapress_premium_countries' );
	    } else {
		// FIXME: no cache in case of service or network error
		return null;
	    }
	}

	return $this->options;
    }

    /**
     * Retrieves licence level of this instance :
     *
     *  - 0 for free version of DyaPress Premium
     *  - 1-10 for activated version of DyaPress Premium at the given level
     *
     * @api
     *
     * @return int|null Licence level of this instance or null on error.
     */
    public function licence_level() {
	if ( empty( $this->options['premium_uuid'] ) || empty( $this->options['expire'] ) || $this->options['expire'] < time() )
	    $this->options();

	return empty( $this->options ) ? null : $this->options['licence_level'];
    }

    /**
     * Determines whether this instance is active, so that API can be used safely.
     *
     * @api
     *
     * @return bool|null True if connexion to DyaPress or Dolibarr is active, false if configuration is not fully done and null on error.
     */
    function is_active() {
	if ( empty( $this->options['premium_uuid'] ) || empty( $this->options['expire'] ) || $this->options['expire'] < time() )
	    $this->options();

	return empty( $this->options ) ? null : $this->options[$this->options['mode'].'_status'] == 'active';
    }

    /**
     * Determines whether this instance is active with purchase authorized.
     *
     * Purchase are authorized if mode is 'live' or if current logged in user has 'manage_options' capability.
     * If mode is 'test', Stripe is used in test mode as well, so purchase are test purchase.
     * To avoid polluting the production ERP only admin users can do purchase and these purchase are redirected to the test ERP.
     *
     * @api
     *
     * @return bool|nul True if connexion to DyaPress or Dolibarr is active with purchase authorized, or null on error.
     */
    public function is_available() {
	if ( empty( $this->options['premium_uuid'] ) || empty( $this->options['expire'] ) || $this->options['expire'] < time() )
	    $this->options();

	return empty( $this->options ) ? null : $this->is_active() && ( $this->options['mode'] == 'live' || current_user_can( 'manage_options' ) );
    }

    /**
     * Retrieves licence level of this instance only if active :
     *
     *  - 1-10 for both an active instance and activated version of DyaPress Premium at the given level
     *  - 0 for both an active instance using free version of DyaPress Premium
     *  - false if instance is not active
     *  - null if an error occurred
     *
     * @api
     *
     * @return int Licence level of this instance.
     */
    public function is_licence_active() {
	if ( empty( $this->options['premium_uuid'] ) || empty( $this->options['expire'] ) || $this->options['expire'] < time() )
	    $this->options();

	return empty( $this->options ) ? null : ( $this->is_active() ? $this->options['licence_level'] : false );
    }

    /**
     * Retrieves payment types from connected ERP.
     *
     * @api
     *
     * @return array|null Payment types array from connected ERP.
     */
    public function get_payment_types() {
	$payment_types = get_transient( 'dyapress_premium_payment_types' );

	if ( false === $payment_types ) {
	    if ( $this->is_active() )
		$payment_types = apply_filters( 'dyapress_premium_payment_types', $this->call_api( 'GET', '/v1/payment_types' ) );

	    if ( !empty( $payment_types ) )
		set_transient( 'dyapress_premium_payment_types', $payment_types );
	}

	return $payment_types;
    }

    /**
     * Retrieves bank accounts from connected ERP.
     *
     * @api
     *
     * @return array|null Bank accounts array from connected ERP.
     */
    public function get_bank_accounts() {
	$bank_accounts = get_transient( 'dyapress_premium_bank_accounts' );

	if ( false === $bank_accounts ) {
	    if ( $this->is_active() )
		$bank_accounts = apply_filters( 'dyapress_premium_bank_accounts', $this->call_api( 'GET', '/v1/bank_accounts' ) );

	    if ( !empty( $bank_accounts ) )
		set_transient( 'dyapress_premium_bank_accounts', $bank_accounts );
	}

	return $bank_accounts;
    }

    /**
     * Retrieves active countries from connected ERP.
     *
     * @api
     *
     * @param string $locale Locale to use for returning countries.
     *
     * @return array|null Active countries array from connected ERP, or null on error.
     */
    public function get_countries( $locale = null ) {
	// in case of short circuiting external API :
	//   return array( (object) array( "id" => 1, "code" => "FR", "name" => "France" ) );
	$countries = get_transient( 'dyapress_premium_countries' );
	if ( empty( $locale ) )
	    $locale = get_locale();

	if ( false === $countries || empty( $countries[$locale] ) ) {
	    if ( $this->is_active() )
		$result = $this->call_api('GET', '/v1/countries?lang='.$locale );

	    if ( !empty( $result ) ) {
		$countries[$locale] = apply_filters( 'dyapress_premium_countries', $result );

		set_transient( 'dyapress_premium_countries', $countries );
	    } else
		return null;
	}

	return $countries[$locale];
    }

    /**
     * Retrieves active european countries from connected ERP.
     *
     * @api
     *
     * @param string $locale Locale to use for returning countries.
     *
     * @return array|null Active european countries from connected ERP, or null on error.
     */
    public function get_european_countries( $locale = null ) {
	if ( empty( $locale ) )
	    $locale = get_locale();

	if ( empty( $this->european_countries ) || empty( $this->european_countries[$locale] ) ) {
	    // This list has been taken from Dolibarr...
	    static $european_country_codes = array(
		'AT',	// Austria
		'BE',	// Belgium
		'BG',	// Bulgaria
		'CY',	// Cyprus
		'CZ',	// Czech republic
		'DE',	// Germany
		'DK',	// Danemark
		'EE',	// Estonia
		'ES',	// Spain
		'FI',	// Finland
		'FR',	// France
		'GB',	// United Kingdom
		'GR',	// Greece
		'HR',	// Croatia
		'NL',	// Holland
		'HU',	// Hungary
		'IE',	// Ireland
		'IM',	// Isle of Man - Included in UK
		'IT',	// Italy
		'LT',	// Lithuania
		'LU',	// Luxembourg
		'LV',	// Latvia
		'MC',	// Monaco - Included in France
		'MT',	// Malta
		//'NO',	// Norway
		'PL',	// Poland
		'PT',	// Portugal
		'RO',	// Romania
		'SE',	// Sweden
		'SK',	// Slovakia
		'SI',	// Slovenia
		'UK',	// United Kingdom
		//'CH',	// Switzerland - No. Swizerland in not in EEC
	    );

	    $countries = $this->get_countries( $locale );

	    if ( isset( $countries ) ) {
		foreach ( $countries as $c ) {
		    if ( in_array( $c->code, $european_country_codes ) )
			$european_countries[$locale][$c->id] = $c->code;
		}
	    } else
		return null;
	}

	return $european_countries[$locale];
    }

    /**
     * Retrieves country object chosen.
     *
     * @api
     *
     * @return object|null Chosen countrie object from connected ERP, or null on error.
     */
    public function country() {
	if ( empty( $this->options['premium_uuid'] ) || empty( $this->options['expire'] ) || $this->options['expire'] < time() )
	    $this->options();

	if ( empty( $this->country ) || $this->country->code != $this->options['country_code'] ) {
	    $countries = $this->get_countries();

	    if ( isset( $countries ) ) {
		foreach ( $countries as $c ) {
		    if ( $c->code == $this->options['country_code'] )
			$this->country = $c;
		}
	    } else
		return null;
	}

	return $this->country;
    }

    /**
     * Retrieves default currency associated to the chosen country.
     *
     * @api
     *
     * @return string|null Default currency associated to country or null if unknown.
     */
    public function default_currency() {
	static $country_currencies = array(
	    'GB' => 'GPB',	// United Kingdom
	    'IM' => 'GPB',	// Isle of Man - Included in UK
	    'NO' => 'NOK',	// Norway
	    'UK' => 'GPB',	// United Kingdom
	    'CH' => 'CHF',	// Switzerland - No. Swizerland in not in EEC
	); // FIXME: to be continued for country not using euro...

	static $default_currency = null;

	if ( empty( $default_currency ) && $this->is_active() ) {
	    if ( empty( $this->options['premium_uuid'] ) || empty( $this->options['expire'] ) || $this->options['expire'] < time() )
		$this->options();

	    if ( isset( $country_currencies[$this->options['country_code']] ) )
		$currency = $country_currencies[$this->options['country_code']];
	    else
		$currency = 'EUR';

	    $default_currency = apply_filters( 'dyapress_premium_default_currency', $currency );
	}

	return $default_currency;
    }

    /**
     * Get product from transient.
     *
     * @api
     *
     * @param object $result Product object.
     * @param string $uuid UUID of product.
     *
     * @return object|false The product object saved in transient, or false if not found.
     */
    public function get_transient_product( $product, $uuid = null ) {
	if ( empty( $uuid ) )
	    $uuid = $this->uuid();

	return get_transient( 'dyapress_premium_product:'.$uuid.':'.$product );
    }

    /**
     * Save product in transient.
     *
     * @api
     *
     * @param object $result Product object.
     * @param string $product Product identification.
     * @param string $uuid UUID of product.
     *
     * @return object|null The product object saved in transient once filter 'dyapress_premium_product' applied, or null on error.
     */
    public function set_transient_product( $result, $product, $uuid = null ) {
	$result = apply_filters( 'dyapress_premium_product', $result, $product );

	if ( empty( $uuid ) )
	    $uuid = $this->uuid();

	if ( !empty( $uuid ) && !empty( $result ) )
	    set_transient( 'dyapress_premium_product:'.$uuid.':'.$product, $result, 4 * HOUR_IN_SECONDS );
	else
	    return null;

	return $result;
    }

    /**
     * Retrieves product from connected ERP.
     *
     * The '$product' parameter can be :
     *
     *  - an integer of no more than 11 digits which is the ID of the product in the connected ERP.
     *  - an integer of 12 or 13 digits which is the barcode (an UPC or EAN13 which will be searched) in the connected ERP.
     *  - a string : the reference of the product in the connected ERP.
     *
     * Any type of value above can furthermore be prefixed with UUID followed by a colon ':' to restrict usage to a given ERP
     * identificatin. Multiple identification are separated by a colon ','.
     *
     * @api
     *
     * @param mixed $product Product identification.
     *
     * @return object|null The product from connected ERP, or null on error.
     */
    public function get_product( $product, $caching = true ) {
	$product = apply_filters( 'dyapress_premium_get_product', $product );

	if ( is_array( $product ) )
	    $product = (object) $product;

	if ( is_object( $product ) )
	    return $product;

	if ( !$this->is_active() )
	    return null;

	$uuid = $this->uuid();
	$product = trim ( $product );

	// check if multiple uuid product mapping are used
	if ( strpos( $product, ':' ) !== false ) {
	    // at this level, $product contains multiple <uuid>:<product> to match for the right UUID
	    if ( preg_match( '/\b'.$uuid.'\s*:\s*([^,;=\s]+)/', $product, $matches ) ) {
		$product = $matches[1];
	    } else
		return null;
	}

	if ( is_numeric( $product ) ) {
	    // a numeric product can id directly the product_id or the reference
	    if ( strlen( $product ) < 12 ) {
		// a numeric of length smaller or equal to 11 digits is a ID of the product
		if ( $caching )
		    $result = $this->get_transient_product( $product, $uuid );
		else
		    $result = false;

		if ( false === $result ) {
		    $result = $this->set_transient_product( $this->call_api( 'GET', '/v1/products/'.$product ), $product, $uuid );

		    // FIXME: barcode transient mapping is not updated to keep multiple products associated in cache
		    // so using get_product() need to be consistant : do not mix real product_id, label or barcode
		    // in order to get maximum speed
		}

		return $result;
	    } else {
		// search for upc, ean13 or other numeric barcode and store the mapping in transient as well
		$key = 'barcode';
	    }
	} else {
	    // search for label and store the mapping in transient as well
	    $key = 'reference';
	}

	if ( $caching )
	    $result = get_transient( 'dyapress_premium_product_mapping:'.$uuid.':'.$product );
	else
	    $result = false;

	if ( false !== $result ) {
	    if ( is_array( $result ) ) {
		$products = array();

		// multiple product are associated, so extract all the corresponding products from transients
		foreach ( $result as $id ) {
		    $products[$id] = $this->get_transient_product( $id, $uuid );

		    if ( false === $products[$id] ) {
			$result = false;
			break;
		    }
		}

		if ( false !== $result )
		    return array_values( $products );
	    } else {
		$result = $this->get_transient_product( $result, $uuid );

		if ( false !== $result )
		    return $result;
	    }
	}

	// now the search must be done again to keep multiple results
	$search_result = $this->call_api( 'GET', '/v1/products?'.$key.'='.rawurlencode( $product ) );
	$ids = array();

	foreach ( $search_result as $result ) {
	    $ids[] = $result->id;

	    $this->set_transient_product( $result, $result->id, $uuid );
	}

	if ( count( $ids ) > 1 ) {
	    set_transient( 'dyapress_premium_product_mapping:'.$uuid.':'.$product, $ids, 24 * HOUR_IN_SECONDS );

	    return $search_result;
	} elseif ( count( $ids ) == 1 ) {
	    set_transient( 'dyapress_premium_product_mapping:'.$uuid.':'.$product, $ids[0], 24 * HOUR_IN_SECONDS );

	    return $search_result[0];
	} else {
	    delete_transient( 'dyapress_premium_product_mapping:'.$uuid.':'.$product );

	    return null;
	}
    }

    function get_product_price( $product, $product_quantity = 1, $thirdparty = null ) {
	if ( !is_object( $product ) )
	    $product = $this->get_product( $product );

	// Allow price computation according to product_quantity to order
	if ( is_object( $product ) ) {
	    $price = $product->price;

	    /*
	     * The following code may be used if price_level per customer is used, as well as mutlprices per level
	     * Currencly the code is not yet used :
	    if ( isset( $user) && isset( $user->price_level ) && isset( $product->multiprices[ $user->price_level ] ) )
		$price = floatval( $product->multiprices[ $user->price_leval ] );
	     */

	    return floatval( apply_filters( 'dyapress_premium_product_price', $price, $product, $product_quantity, $thirdparty ) );
	} else
	    return null;
    }

    function get_total_price( $product, $product_quantity, $thirdparty = null ) {
	$price = $this->get_product_price( $product, $product_quantity, $thirdparty );

	if ( isset( $price ) )
	    $price = $product_quantity * $price;

	return $price;
    }

    function get_product_vat( $product, $thirdparty = null ) {
	if ( !is_object( $product ) )
	    $product = $this->get_product( $product );

	// Allow VAT correction according to country of thirdparty
	if ( is_object( $product ) ) {
	    $vat = $product->vat;

	    if ( !empty( $thirdparty ) ) {
		if ( !$this->is_vat_applicable( $thirdparty ) )
		    $vat = 0;
	    }

	    return floatval( apply_filters( 'dyapress_premium_product_vat', $vat, $thirdparty ) );
	} else
	    return null;
    }

    function get_product_price_vat_excl( $product, $product_quantity = 1, $thirdparty = null ) {
	return $this->get_product_price( $product, $product_quantity, $thirdparty );
    }

    function get_total_price_vat_excl( $product, $product_quantity, $thirdparty = null ) {
	return $this->get_total_price( $product, $product_quantity, $thirdparty );
    }

    function get_total_price_vat_incl( $product, $product_quantity, $thirdparty = null ) {
	$price = $this->get_total_price( $product, $product_quantity, $thirdparty );
	$vat = $this->get_product_vat( $product, $thirdparty );

	if ( isset( $price ) && isset( $vat ) )
	    $price = $price * (100 + $vat ) / 100;
	else
	    $price = null;

	return $price;
    }

    private function _complete_user( $wp_user, $cache_thirdparties = null ) {
	if ( is_object( $wp_user ) && $wp_user->ID > 0 ) {
	    $user = array(
		'firstname' => $wp_user->user_firstname,
		'lastname' => $wp_user->user_lastname,
		'email' => $wp_user->user_email,
		'login' => $wp_user->user_login,
		'display_name' => $wp_user->display_name,
		'ID' => $wp_user->ID,
	    );

	    if ( $this->is_active() ) {
		$uuid = $this->uuid();
		$meta = get_user_meta( $wp_user->ID );

		if ( isset( $meta['dyapress_premium_thirdparties_expire:'.$uuid] ) && isset( $meta['dyapress_premium_thirdparties_expire:'.$uuid][0] ) ) {
		    $user['thirdparties_expire'] = (int)$meta['dyapress_premium_thirdparties_expire:'.$uuid][0];
		} else
		    $user['thirdparties_expire'] = 0;
		//print_r("thirdparties_expire for user ID=".$user['ID']." (".$user['login'].") is ".date("Y-m-d H:i:s", $user['thirdparties_expire'])." and now is ".date("Y-m-d H:i:s"));

		// FIXME: cache_thirdparties exactly equal to false means no cache update (admin tab orders may be slow otherwise)
		if ( $cache_thirdparties !== false && $user['thirdparties_expire'] < time() ) {
		    // Thirdparties have expired, so get fresh copy cached for 1 hour
		    $result = $this->call_api('GET', '/v1/thirdparties?email='.urlencode( $wp_user->user_email ) );

		    if ( isset( $result ) ) {
			$thirdparties = array();
			if ( is_array( $result ) ) {
			    foreach ( $result as $t ) {
				$thirdparties[$t->id] = $this->validate_external_thirdparty( $t );
				$thirdparties[$t->id]->ID = $wp_user->ID;
			    }
			}

			if ( !isset( $cache_thirdparties ) )
			    $cache_thirdparties = apply_filters( 'dyapress_premium_cache_thirdparties', 3600 );

			$user['thirdparties'] = $thirdparties;
			$user['thirdparties_expire'] = $cache_thirdparties + time();

			update_user_meta( $wp_user->ID, 'dyapress_premium_thirdparties:'.$uuid, $thirdparties );
			update_user_meta( $wp_user->ID, 'dyapress_premium_thirdparties_expire:'.$uuid, $user['thirdparties_expire'] );
		    }
		} else {
		    if ( isset( $meta['dyapress_premium_thirdparties:'.$uuid] ) && isset( $meta['dyapress_premium_thirdparties:'.$uuid][0] ) )
			$user['thirdparties'] = maybe_unserialize( $meta['dyapress_premium_thirdparties:'.$uuid][0] );
		    else
			$user['thirdparties'] = array();
		}

		return apply_filters( 'dyapress_premium_user', (object)$user );
	    }
	}

	return null;
    }

    function has_user_field( $name ) {
	if ( in_array( $name, array( 'firstname', 'lastname', 'email', 'login', 'display_name', 'ID' ) ) )
	    return true;

	return isset( $this->user_meta_fields[$name] );
    }

    function register_user_field( $name, $validation, $thirdparty_field = null ) {
	if ( in_array( $name, array( 'firstname', 'lastname', 'email', 'login', 'display_name', 'ID' ) ) )
	    return false;

	if ( $validation == 'int' || $validation == 'str' || is_callable( $validation ) ) {
	    $this->user_meta_fields[$name] = array (
		'thirdparty_field' => isset( $thirdparty_field ) ? $thirdparty_field : $name,
		'validation' => $validation,
	    );

	    return true;
	} else
	    return false;
    }

    function deregister_user_field( $name ) {
	if ( isset( $this->user_meta_fields[$name] ) ) {
	    unset( $this->user_meta_fields[$name] );

	    return true;
	} else
	    return false;
    }

    function get_current_user( $cache_thirdparties = null ) {
	if ( is_user_logged_in() )
	    $wp_user = wp_get_current_user();

	return self::_complete_user( $wp_user, $cache_thirdparties );
    }

    function get_current_thirdparty( $cache_thirdparties = null ) {
	$current_user = $this->get_current_user( $cache_thirdparties );

	if ( !empty( $current_user ) ) {
	    if ( is_array( $current_user->thirdparties ) ) {
		if ( !empty( $current_user->thirdparty_id ) && isset( $current_user->thirdparties[$current_user->thirdparty_id] ) )
		    return $current_user->thirdparties[$current_user->thirdparty_id];
		elseif ( count( $current_user->thirdparties ) > 0 )
		    return array_values( $current_user->thirdparties )[0];
	    }
	}

	return null;
    }

    // search for user and return false if not found, null if error or another user exists but with different data
    // FIXME: thirdparty_id is set only if param is an array
    function get_user( $param, $cache_thirdparties = null ) {
	if ( $this->is_active() && !empty( $param ) ) {
	    if ( is_object( $param ) )
		$param = (array)$param;

	    if ( is_array( $param ) ) {
		if ( isset( $param['email'] ) )
		    $user_id = email_exists( $param['email'] );
		elseif ( isset( $param['ID'] ) )
		    $user_id = (int)$param['ID'];
		else
		    return null;

		if ( !$user_id )
		    return false;

		$existing_user = self::_complete_user( get_userdata( $user_id ), $cache_thirdparties );

		// check if existing user already exists
		if ( isset( $existing_user ) ) {
		    $current_user = $this->get_current_user();

		    if ( !empty( $current_user ) && $existing_user->ID == $current_user->ID ) {
			// special case for logged in user : only firstname, lastname and email are verified with current user
			foreach ( array( 'firstname', 'lastname', 'email' ) as $k ) {
			    if ( isset( $param[$k] ) && $param[$k] != $existing_user->$k || $existing_user->$k != $current_user->$k )
				return null;
			}

			// check if existing user match a thirdparty, matching_user() will return false if no user is found
			// or return existing user with thirdparty_id set to the corresponding thirdparty if found
			// FIXME: this approach will avoid creating an already existing thirdparty
			$matching_user = $this->matching_user( $param, $existing_user );
			if ( !empty( $matching_user ) )
			    return $matching_user;

			// user is logged in and a new thirdparty can be created safely
			if ( ( $uuid = $this->uuid() ) ) {
			    if ( empty( $param['thirdparty_id'] ) ) {
				$result = $this->call_api('POST', '/v1/thirdparties', $this->validate_param_to_external_thirdparty( $param ) );
				if ( isset( $result ) ) {
				    $existing_user->thirdparties[$result->id] = $this->validate_external_thirdparty( $result );
				    $existing_user->thirdparties[$result->id]->ID = $user_id;

				    if ( !isset( $cache_thirdparties ) )
					$cache_thirdparties = apply_filters( 'dyapress_premium_cache_thirdparties', 3600 );

				    $existing_user->thirdparty_id = $result->id;
				    $existing_user->thirdparty_expire = $cache_thirdparties + time();

				    update_user_meta( $user_id, 'dyapress_premium_thirdparties:'.$uuid, $existing_user->thirdparties );
				    update_user_meta( $user_id, 'dyapress_premium_thirdparties_expire:'.$uuid, $existing_user->thirdparties_expire );

				    do_action( 'dyapress_premium_thirdparty_created', $existing_user, $param );

				    return $existing_user;
				}
			    }
			}
		    } else {
			// there are no user connected, but we need to check if the existing user fully match
			return $this->matching_user( $param, $existing_user );
		    }

		    return null;
		} else
		    return false;
	    } else {
		if ( !is_numeric( $param ) )
		    $param = email_exists( $param );

		return $param ? self::_complete_user( get_userdata( $param ), $cache_thirdparties ) : false;
	    }
	}

	return null;
    }

    private function matching_user( $param, &$existing_user ) {
	$existing_user_match = true;

	if ( empty( $param ) )
	    return null;
	elseif ( is_object( $param ) )
	    $param = (array)$param;

	// Step 1 : check if existing user match firstname, lastname and email
	foreach ( array( 'firstname', 'lastname', 'email' ) as $k ) {
	    if ( isset( $param[$k] ) && $param[$k] != $existing_user->$k )
		return null;
	}

	// Step 2 : check if provided thirdparty_id matchs
	if ( !empty( $param['thirdparty_id'] ) ) {
	    if ( is_array( $existing_user->thirdparties ) )
		$thirdparty = $existing_user->thirdparties[ (int)$param['thirdparty_id'] ];

	    if ( !empty( $thirdparty ) ) {
		if ( $thirdparty->thirdparty_id != (int)$param['thirdparty_id'] )
		    return null;
		else {
		    foreach ( $thirdparty as $k => $v ) {
			if ( isset( $param[$k] ) && sanitize_text_field( $param[$k] ) != sanitize_text_field( $v ) ) {
			    $existing_user_match = false;
			    break;
			}
		    }
		    if ( $existing_user_match ) {
			// keep in mind which thirdparty match
			$existing_user->thirdparty_id = $thirdparty->thirdparty_id;

			return $existing_user;
		    } else
			return null;
		}
	    } else
		return null;
	}

	// Step 3 : loop over thirdparties to find a thirdparty that match
	if ( !empty( $existing_user->thirdparties ) && is_array( $existing_user->thirdparties ) ) {
	    foreach ( $existing_user->thirdparties as $thirdparty ) {
		$existing_user_match = true;

		foreach ( $thirdparty as $k => $v ) {
		    if ( isset( $param[$k] ) && sanitize_text_field( $param[$k] ) != sanitize_text_field( $v ) ) {
			$existing_user_match = false;
			break;
		    }
		}

		if ( $existing_user_match ) {
		    // keep in mind which thirdparty match
		    $existing_user->thirdparty_id = $thirdparty->thirdparty_id;

		    return $existing_user;
		}
	    }
	}

	return false;
    }

    function validate_user( $user ) {
	if ( !empty( $user ) ) {
	    if ( is_array( $user ) )
		$user = (object)$user;

	    $validated_user = array();

	    foreach ( array( 'thirdparty_id', 'thirdparty_contact_id' ) as $k )
		if ( isset( $user->$k ) && (int)$user->$k > 0 )
		    $validated_user[$k] = (int)$user->$k;

	    if ( !empty( $user->ID ) && (int)$user->ID > 0 )
		$validated_user['ID'] = (int)$user->ID;
	    if ( !empty( $user->email ) && filter_var( $user->email, FILTER_VALIDATE_EMAIL ) )
		$validated_user['email'] = sanitize_email( $user->email );
	    if ( !empty( $user->firstname ) )
		$validated_user['firstname'] = sanitize_text_field( $user->firstname );
	    if ( !empty( $user->lastname ) )
		$validated_user['lastname'] = sanitize_text_field( $user->lastname );

	    foreach ( $this->user_meta_fields as $k => $meta ) {
		$validation = $meta['validation'];

		if ( isset( $user->$k ) ) {
		    if ( $validation == 'int' )
			$validated_user[$k] = (int)$user->$k;
		    elseif ( $validation == 'str' )
			$validated_user[$k] = sanitize_text_field( $user->$k );
		    elseif ( is_callable( $validation ) )
			$validated_user[$k] = call_user_func( $validation, $user->$k );
		}
	    }

	    return (object)$validated_user;
	}

	return null;
    }

    function validate_param_to_external_thirdparty( $user ) {
	if ( !empty( $user ) ) {
	    if ( is_array( $user ) )
		$user = (object)$user;

	    $thirdparty = array();

	    foreach ( array( 'thirdparty_id' => 'id', 'thirdparty_contact_id' => 'contact_id' ) as $k => $v )
		if ( isset( $user->$k ) && (int)$user->$k > 0 )
		    $thirdparty[$v] = (int)$user->$k;

	    if ( isset( $user->email ) && filter_var( $user->email, FILTER_VALIDATE_EMAIL ) )
		$thirdparty['contact_email'] = $user->email;
	    if ( !empty( $user->firstname ) )
		$thirdparty['contact_firstname'] = wp_strip_all_tags( $user->firstname );
	    if ( !empty( $user->lastname ) )
		$thirdparty['contact_lastname'] = wp_strip_all_tags( $user->lastname );

	    foreach ( $this->user_meta_fields as $k => $meta ) {
		$thirdparty_field = $meta['thirdparty_field'];
		$validation = $meta['validation'];

		if ( !empty( $thirdparty_field ) ) {
		    if ( isset( $user->$k ) ) {
			if ( $validation == 'int' )
			    $thirdparty[$thirdparty_field] = (int)$user->$k;
			elseif ( $validation == 'str' )
			    $thirdparty[$thirdparty_field] = sanitize_text_field( $user->$k );
			elseif ( is_callable( $validation ) )
			    $thirdparty[$thirdparty_field] = call_user_func( $validation, $user->$k );
		    }
		}
	    }

	    return (object)$thirdparty;
	}

	return null;
    }

    function validate_external_thirdparty( $thirdparty ) {
	if ( !empty( $thirdparty ) ) {
	    if ( is_array( $thirdparty ) )
		$thirdparty = (object)$thirdparty;

	    $user = array();

	    foreach ( array(
		'id' => 'thirdparty_id', 'contact_id' => 'thirdparty_contact_id',
		'contact_email' => 'email', 'contact_firstname' => 'firstname', 'contact_lastname' => 'lastname' ) as $v => $k )
		if ( isset( $thirdparty->$v ) )
		    $user[$k] = $thirdparty->$v;

	    foreach ( $this->user_meta_fields as $k => $meta ) {
		$thirdparty_field = $meta['thirdparty_field'];

		if ( !empty( $thirdparty_field ) ) {
		    if ( isset( $thirdparty->$thirdparty_field ) )
			$user[$k] = $thirdparty->$thirdparty_field;
		}
	    }

	    return (object)$user;
	}

	return null;
    }

    function validate_external_thirdparties( $thirdparties ) {
	if ( is_array( $thirdparties ) ) {
	    $users = array();

	    foreach ( $thirdparties as $thirdparty ) {
		$user = $this->validate_thirdparty_to_user( $thirdparty );

		if ( isset( $user ) )
		    $users[] = $user;
		else
		    return null;
	    }
	}

	return null;
    }

    // FIXME: create_thirdparty will both create a new customer (Dolibarr) and a new WP user
    // FIXME: assign_thirdparty will assign an existing (Dolibarr) customer to a new WP user
    // FIXME: search_thirdparty will search for existing customers (Dolbarr)
 
    function create_thirdparty( $user, $role = false, $cache_thirdparties = null ) {
	if ( $this->is_active() ) {
	    // email is the minimal field required to create a WP account
	    $created_user = $this->create_user( $user, $role );

	    if ( !empty( $created_user ) && !empty( $created_user->ID ) ) {
		$result = $this->call_api('GET', '/v1/thirdparties?email='.urlencode( $created_user->email ) );

		if ( isset( $result ) ) {
		    $thirdparties = array();
		    if ( is_array( $result ) ) {
			foreach ( $result as $t ) {
			    $thirdparties[$t->id] = $this->validate_external_thirdparty( $t );
			    $thirdparties[$t->id]->ID = $wp_user->ID;
			}
		    }

		    if ( !isset( $cache_thirdparties ) )
			$cache_thirdparties = apply_filters( 'dyapress_premium_cache_thirdparties', 3600 );

		    $created_user->thirdparties = $thirdparties;
		    $created_user->thirdparties_expire = $cache_thirdparties + time();
		}

		$existing_user = $this->matching_user( $user, $created_user );

		if ( empty( $existing_user ) ) {
		    $result = $this->call_api('POST', '/v1/thirdparties', $this->validate_param_to_external_thirdparty( $user ) );

		    if ( isset( $result ) ) {
			$created_user->thirdparty_id = (int)$result->id;

			if ( empty($created_user->thirdparties) )
			    $created_user->thirdparties = array( (int)$result->id => $result );
			else
			    $created_user->thirdparties[(int)$result->id] = $result;

			if ( !isset( $cache_thirdparties ) )
			    $cache_thirdparties = apply_filters( 'dyapress_premium_cache_thirdparties', 3600 );

			$created_user->thirdparties_expire = $cache_thirdparties + time();
		    }
		} else {
		    $created_user->thirdparty_id = $existing_user->thirdparty_id;
		}

		$uuid = $this->uuid();
		update_user_meta( $created_user->ID, 'dyapress_premium_thirdparties:'.$uuid, $created_user->thirdparties );
		update_user_meta( $created_user->ID, 'dyapress_premium_thirdparties_expire:'.$uuid, $created_user->thirdparties_expire );

		do_action( 'dyapress_premium_thirdparty_created', $created_user, $param );

		return $created_user;
	    }
	}

	return null;
    }

    function search_thirdparties_by_email( $user ) {
	if ( !empty( $user ) ) {
	    // search for customers based on email only, beware to verify email address before creating an user in WP
	    if ( is_object( $user ) )
		$user = $user->email;
	    elseif ( is_array( $user ) )
		$user = $user['email'];
	    else
		$user = null;

	    if ( isset( $user ) && filter_var( $user, FILTER_VALIDATE_EMAIL ) ) {
		$result = $this->call_api('GET', '/v1/thirdparties?email='.urlencode( $user ) );

		if ( isset( $result ) )
		    return $this->validate_external_thirdparties( $result );
	    }
	}

	return null;
    }

    function create_user( $user, $role = false, $cache_thirdparties = null ) {
	if ( !empty( $user ) ) {
	    if ( is_array( $user ) )
		$user = (object)$user;

	    // create a user using email, password will be unknown, so the user must choose it at some stage
	    // login is optional and will be guessed from email if not given
	    // role is optional as well
	    if ( isset( $user->email ) && filter_var( $user->email, FILTER_VALIDATE_EMAIL ) )
		$email_user_id = email_exists( $user->email );

	    if ( $email_user_id == false ) {
		if ( isset( $user->login ) )
		    $login = $user->login;
		else
		    $login = preg_replace( '/@.*/', '', $user->email );

		while ( ( $login_user_id = username_exists( $login ) ) ) {
		    if ( preg_match( '/(.*?)(\d+)$/', $login, $matches ) )
			$login = $matches[1] . ( (int)$matches[2] + 1 );
		    else
			$login = $login . '2';
		}

		$user_id = wp_create_user( $login, wp_generate_password(), $user->email );

		if ( !is_wp_error( $user_id ) && $user_id > 0 ) {
		    $update_user = array( 'ID' => $user_id, 'show_admin_bar_front' => 'false' );

		    if ( isset( $user->firstname ) )
			$update_user['first_name'] = $user->firstname;
		    if ( isset( $user->lastname ) )
			$update_user['last_name'] = $user->lastname;

		    wp_update_user( $update_user );

		    $wp_user = get_userdata( $user_id );
		    if ( !empty( $role ) )
			$wp_user->set_role( $role );

		    $created_user = self::_complete_user( $wp_user, $cache_thirdparties );

		    do_action( 'dyapress_premium_user_created', $created_user, $user );

		    return $created_user;
		}
	    } else
		return self::_complete_user( get_userdata( $email_user_id ), $cache_thirdparties );
	}

	return null;
    }

    // return sanitized $email if email need to been changed, false if email is not changed, null on error
    function update_user( $user, $param ) {
	if ( !empty( $param ) ) {
	    if ( is_object( $param ) )
		$param = (array)$param;

	    if ( is_array( $param ) ) {
		if ( !empty( $param['firstname'] ) )
		    $firstname = sanitize_text_field( $param['firstname'] );
		if ( !empty( $param['lastname'] ) )
		    $lastname = sanitize_text_field( $param['lastname'] );
		if ( !empty( $param['email'] ) )
		    $email = sanitize_text_field( $param['email'] );

		$update_user = array();
		if ( isset( $firstname ) && $user->firstname != $firstname )
		    $update_user['first_name'] = $firstname;
		if ( isset( $lastname ) && $user->lastname != $lastname )
		    $update_user['last_name'] = $lastname;
		if ( count( $update_user ) > 0 ) {
		    $update_user['ID'] = $user->ID;

		    $user_id = wp_update_user( $update_user );

		    if ( !is_wp_error( $user_id ) && $user_id > 0 ) {
			$user->firstname = $firstname;
			$user->lastname = $lastname;
		    } else
			return null;
		}

		return !empty( $email ) && $user->email != $email ? $email : '';
	    }
	}

	return null;
    }

    function update_user_email ( $user, $email, $cache_thirdparties = null ) {
	if ( $this->is_active() && !empty( $user ) && isset( $email ) ) {
	    if ( is_object ( $email ) )
		$email = $email->email;
	    elseif ( is_array( $email ) )
		$email = $email['email'];

	    $email = sanitize_email( $email );

	    if ( is_array( $user ) )
		$user = (object)$user;

	    if ( $user->email != $email ) {
		$update_user = array(
		    'ID' => $user->ID,
		    'user_email' => $email,
		);

		$wp_user = wp_update_user( $update_user );

		if ( !is_wp_error( $user_id ) && $user_id > 0 ) {
		    // TODO: update all thirdparty contacts at once

		    update_user_meta( $user_id, 'dyapress_premium_thirdparties_expire:'.$this->uuid(), 0 );

		    // force fetching of all thirdparties
		    return self::_complete_user( get_userdata( $user_id ), $cache_thirdparties );
		} else
		    return null;
	    }

	    return $user;
	}

	return null;
    }

    function update_thirdparty( $user, $param, $cache_thirdparties = null ) {
	if ( $this->active() && !empty( $user ) ) {
	    if ( is_array( $user ) )
		$user = (object)$user;


	    if ( !empty( $uuid ) && !empty( $param['thirdparty_id'] ) && !empty( $user->thirdparties ) ) {
		// update a specific thirdparty of user
		$param['email'] = $user->email;

		$result = $this->call_api('POST', '/v1/thirdparties', $this->validate_param_to_external_thirdparty( $param ) );
		if ( isset( $result ) ) {
		    $user->thirdparties[(int)$result->id] = $result;

		    if ( !isset( $cache_thirdparties ) )
			$cache_thirdparties = apply_filters( 'dyapress_premium_cache_thirdparties', 3600 );

		    $user->thirdparties_expire = $cache_thirdparties + time();

		    $uuid = $this->uuid();

		    update_user_meta( $user->ID, 'dyapress_premium_thirdparties:'.$uuid, $user->thirdparties );
		    update_user_meta( $user->ID, 'dyapress_premium_thirdparties_expire:'.$uuid, $user->thirdparties_expire );

		    do_action( 'dyapress_premium_thirdparty_updated', $user, $param );

		    return $user;
		}
	    }
	}

	return null;
    }

    function get_proposal( $id ) {
	if ( $this->active() && $id > 0 ) {
	    $result = $this->call_api('GET', '/v1/proposals/'.$id );

	    if ( isset( $result ) )
		return apply_filters( 'dyapress_premium_proposal', $result );
	}

	return null;
    }

    function get_proposals( $args = false ) {
	if ( $this->is_active() ) {
	    $params = array();

	    if ( is_array( $args) ) {
		if ( !empty( $args['customer'] ) ) {
		    // extract thirdparty_id from customer email
		    if ( $user_id = email_exists( $args['customer'] ) ) {
			$user = $this->get_user( $user_id );
			$params['thirdparty_ids'] = join( ',', array_keys( $user->thirdparties ) );

			// make sure thirdparty_ids is defined
			if ( empty( $params['thirdparty_ids'] ) )
			    return false;
		    } else
			return false;
		}
		if ( !empty( $args['customer_id'] ) ) {
		    if ( $user_id = (int)$args['customer_id'] ) {
			$user = $this->get_user( $user_id );
			$params['thirdparty_ids'] = join( ',', array_keys( $user->thirdparties ) );

			// make sure thirdparty_ids is defined
			if ( empty( $params['thirdparty_ids'] ) )
			    return false;
		    } else
			return false;
		}
		if ( !empty( $args['status'] ) ) {
		    if ( is_array( $args['status'] ) )
			$params['status'] = join( ',', $args['status'] );
		    elseif ( preg_match( '/^[\w\s,]+$/', $args['status'] ) )
			$params['status'] = preg_replace( '/[^\w,]/', '', $args['status'] );
		}
	    } elseif ( $args )
		return null;

	    $query = '';
	    if ( count( $params ) )
		$query = '?'.http_build_query( $params, '', '&' );

	    $result = $this->call_api('GET', '/v1/proposals'.$query);

	    if ( isset( $result ) )
		return apply_filters( 'dyapress_premium_proposals', $result );
	}

	return null;
    }

    function get_invoice( $id ) {
	if ( $this->is_active() && $id > 0 ) {
	    $result = $this->call_api('GET', '/v1/invoices/'.$id );

	    if ( isset( $result ) )
		return apply_filters( 'dyapress_premium_invoice', $result );
	}

	return null;
    }

    function get_invoices( $args = false ) {
	if ( $this->is_active() ) {
	    $params = array();

	    if ( is_array( $args) ) {
		if ( isset( $args['customer'] ) ) {
		    // extract thirdparty_id from customer email
		    if ( $user_id = email_exists( $args['customer'] ) ) {
			$user = $this->get_user( $user_id );
			$params['thirdparty_ids'] = join( ',', array_keys( $user->thirdparties ) );

			// make sure thirdparty_ids is defined
			if ( empty( $params['thirdparty_ids'] ) )
			    return false;
		    } else
			return false;
		}
		if ( isset( $args['customer_id'] ) ) {
		    if ( $user_id = (int)$args['customer_id'] ) {
			$user = $this->get_user( $user_id );
			$params['thirdparty_ids'] = join( ',', array_keys( $user->thirdparties ) );

			// make sure thirdparty_ids is defined
			if ( empty( $params['thirdparty_ids'] ) )
			    return false;
		    } else
			return false;
		}
	    } elseif ( $args )
		return null;

	    $query = '';
	    if ( count( $params ) )
		$query = '?'.http_build_query( $params, '', '&' );

	    $result = $this->call_api('GET', '/v1/invoices'.$query);

	    if ( isset( $result ) )
		return apply_filters( 'dyapress_premium_invoices', $result );
	}

	return null;
    }

    function download_document( $doc ) {
	if ( $this->is_active() && $doc ) {
	    $result = $this->call_api('GET', '/v1/documents/'.$doc );

	    if ( isset( $result ) )
		return apply_filters( 'dyapress_premium_document', $result );
	}

	return null;
    }

    function build_document( $doc, $template, $locale ) {
	if ( $this->is_active() && $doc ) {
	    if ( empty( $template ) ) {
		if ( preg_match( '/^commande\//', $doc ) )
		    $template = 'einstein';
		elseif ( preg_match( '/^facture\//', $doc ) )
		    $template = 'crabe';
		elseif ( preg_match( '/^propale\//', $doc ) )
		    $template = 'azur';
	    }

	    if ( empty( $locale ) )
		$locale = get_locale();

	    $result = $this->call_api('PUT', '/v1/documents/'.$doc, array( 'lang' => $locale, 'template' => $template ) );

	    if ( isset( $result ) )
		return apply_filters( 'dyapress_premium_document', $result );
	}

	return null;
    }

    /**
     * Create an order of lines of product in default order state "draft" using given currency and inital order meta.
     *
     * An order must have at least of line of product.
     * The order is not associated to any user or thirdparty yet. The order will be associated to a thirdparty with confirm_order().
     *
     * @api
     *
     * @param mixed $order object Order object or array, or id.
     * @param array $lines Array of product to order.
     * @param string|null $currency Currency code of the order of up to 4 chars, or null for default currency.
     * @param array|null $order_meta Meta associated to the order.
     *
     * @return object Order updated.
     */
    function create_order( $lines, $currency = null, $order_meta = null ) {
	if ( $this->is_active() && sizeof( $lines ) > 0 && ( empty( $order_meta ) || is_array( $order_meta ) ) ) {
	    if ( empty( $currency ) )
		$currency = $this->default_currency();

	    $order = array(
		'currency' => $currency,
		'order_lines' => $lines,
		'order_meta' => $order_meta,
	    );
	    if ( is_user_logged_in() )
		$order['user_id'] = get_current_user_id();

	    $order = apply_filters( 'dyapress_premium_create_order', $order );

	    // FIXME: make sure order_meta is defined or null
	    if ( empty( $order['order_meta'] ) || count( $order['order_meta'] ) == 0 )
		$order['order_meta'] = null;

	    $result = $this->call_api('POST', '/v1/orders', $order );

	    if ( isset( $result ) )
		return apply_filters( 'dyapress_premium_order_created', $result );
	}

	return null;
    }

    /**
     * Update an existing order.
     *
     * @api
     *
     * @param mixed $order object Order object or array, or id.
     * @param array $lines Array of product to order.
     * @param string|null $currency Currency code of the order of up to 4 chars, or null for default currency.
     * @param array|null $order_meta Meta associated to the order.
     *
     * @return object Order updated.
     */
    public function update_order( $order, $lines, $currency = null, $order_meta = null ) {
	if ( $this->is_active() ) {
	    $id = $this->get_order_id( $order );

	    if ( !empty( $id ) ) {
		if ( empty( $currency ) )
		    $currency = $this->default_currency();

		$order = apply_filters( 'dyapress_premium_update_order', array(
		    'currency' => $currency,
		    'order_lines' => $lines,
		    'order_meta' => $order_meta,
		), $id );

		// FIXME: make sure order_meta is set or null
		if ( empty( $order['order_meta'] ) || count( $order['order_meta'] ) == 0 )
		    $order['order_meta'] = null;

		$result = $this->call_api('PUT', '/v1/orders/'.$id, $order );

		if ( isset( $result ) )
		    return apply_filters( 'dyapress_premium_order_updated', $result );
	    }
	}

	return null;
    }

    /**
     * Retrieve Id of order.
     *
     * @api
     *
     * @param $order Order object or array, or id.
     *
     * @return int Id of order, or 0 on any error.
     */
    function get_order_id ( $order ) {
	if ( isset( $order ) ) {
	    if ( is_object( $order ) )
		$id = (int)$order->id;
	    else if ( is_array( $order ) )
		$id = (int)$order['id'];
	    else if ( (int)$order > 0 )
		$id = (int)$order;
	}

	return !empty( $id ) ? $id : 0;
    }

    function confirm_order( $order, $user ) {
	if ( $this->is_active() && !empty( $user ) ) {
	    // confirm an order by assigning a customer and start processing it
	    $id = $this->get_order_id( $order );

	    if ( !empty( $id ) ) {
		if ( is_array( $user ) )
		    $user = (object)$user;

		if ( !empty( $user->thirdparty_id ) && (int)$user->thirdparty_id > 0 ) {
		    $result = $this->call_api('POST', '/v1/orders/'.$id.'/confirm', array(
			'thirdparty_id' => (int)$user->thirdparty_id,
			'user_id' => (int)$user->ID,
		    ) );

		    if ( isset( $result ) )
			return apply_filters( 'dyapress_premium_order_confirmed', $result );
		}
	    }
	}

	return null;
    }

    function manage_orders( $orders ) {
	if ( $this->is_active() && is_array( $orders ) ) {
	    $args = array( 'orders' => array() );

	    foreach ( $orders as $order ) {
		$id = $this->get_order_id( $order );

		if ( !empty( $id ) )
		    $args['orders'][] = $id;
		else
		    return null;
	    }

	    $result = $this->call_api( 'POST', '/v1/orders/manage', $args );

	    if ( isset( $result ) )
		return $result;
	}

	return null;
    }

    function prepare_orders( $orders ) {
	if ( $this->is_active() && is_array( $orders ) ) {
	    $args = array( 'orders' => array() );

	    foreach ( $orders as $order ) {
		$id = $this->get_order_id( $order );

		if ( !empty( $id ) )
		    $args['orders'][] = $id;
		else
		    return null;
	    }

	    $result = $this->call_api( 'POST', '/v1/orders/prepare', $args );

	    if ( isset( $result ) )
		return $result;
	}

	return null;
    }

    function complete_orders( $orders ) {
	if ( $this->is_active() && is_array( $orders ) ) {
	    $args = array( 'orders' => array() );

	    foreach ( $orders as $order ) {
		$id = $this->get_order_id( $order );

		if ( !empty( $id ) )
		    $args['orders'][] = $id;
		else
		    return null;
	    }

	    $result = $this->call_api( 'POST', '/v1/orders/complete', $args );

	    if ( isset( $result ) )
		return $result;
	}

	return null;
    }

    function prepare_order( $order ) {
	// prepare an order : capture payment (if a payment has been made) and update order_state to the following values :
	//   processing: the order only contains services and can be processed immediately (order_state may already be processing)
	//   preparing: all products are available in stock, a delivery can be made
	//   backorder: some products are not available in stock
	//   interrupted: a problem occurred with DyaPress or Dolibarr (some product deleted, no response, ...
	//   cancelled : a problem occurred with Stripe to capture payment
	//
	// an order can be prepared multiple times
	// if an order is not paid the order_state will be unchanged
	//
	// TODO: a new function will be made available to create a delivery just after calling prepare_order()
	//
	// Return an object which contains :
	//   order_state: the order state of the order (same as $order->order_state)
	//   products: the list of physical products of the order (extracted from order_lines)
	//   services: the list of services of the order (extracted from order_lines)
	if ( $this->is_active() ) {
	    $id = $this->get_order_id( $order );

	    if ( !empty( $id ) )
		$result = $this->call_api('POST', '/v1/orders/'.$id.'/prepare' );

	    if ( isset( $result ) )
		return apply_filters( 'dyapress_premium_order_prepared', $result );
	}

	return null;
    }

    function complete_order( $order ) {
	// complete an order by indicating the job is done, so order can be closed and invoice prepared
	// the order can only be completed once
	if ( $this->is_active() ) {
	    $id = $this->get_order_id( $order );

	    if ( !empty( $id ) )
		$result = $this->call_api('POST', '/v1/orders/'.$id.'/complete' );

	    if ( isset( $result ) )
		return apply_filters( 'dyapress_premium_order_completed', $result );
	}

	return null;
    }

    function get_order( $order ) {
	if ( $this->is_active() ) {
	    $id = $this->get_order_id( $order );

	    if ( !empty( $id ) )
		$result = $this->call_api('GET', '/v1/orders/'.$id );

	    if ( isset( $result ) )
		return apply_filters( 'dyapress_premium_order', $result );
	}

	return null;
    }

    function get_orders( $args = false ) {
	if ( $this->is_active() ) {
	    $params = array();

	    if ( is_array( $args) ) {
		if ( isset( $args['customer'] ) ) {
		    // extract thirdparty_id from customer email
		    if ( $user_id = email_exists( $args['customer'] ) )
			$params['user_id'] = $user_id;
		    else
			return false;
		}
		if ( isset( $args['customer_id'] ) ) {
		    if ( $user_id = (int)$args['customer_id'] )
			$params['user_id'] = $user_id;
		    else
			return false;
		}

		if ( isset( $args['order_state'] ) )
		    $params['order_state'] = sanitize_key( $args['order_state'] );

		if ( isset( $args['currency'] ) )
		    $params['currency'] = strtoupper( sanitize_key( $args['currency'] ) );

		if ( isset( $args['return'] ) && $args['return'] = 'ids' )
		    $params['return'] = 'ids';

		if ( isset( $args['date_created'] ) ) {
		    // FIXME: normalize as WooCommerce date and extract date_created_min and date_created_max in unix timestamp
		}

		if ( isset( $args['offset'] ) )
		    $params['offset'] = (int)$args['offset'] > 0 ? (int)$args['offset'] : 0;
		if ( isset( $args['paged'] ) )
		    $params['paged'] = (int)$args['paged'] > 0 ? (int)$args['paged'] : 1;
		if ( isset( $args['limit'] ) )
		    $params['limit'] = (int)$args['limit'] > 0 ? (int)$args['limit'] : (int)get_option( 'posts_per_page' );
		else
		    $params['limit'] = (int)get_option( 'posts_per_page' );
		if ( isset( $args['paginate'] ) )
		    $params['paginate'] = (int)$args['paginate'] > 1 ? (int)$args['paginate'] : 1;

		if ( isset( $args['orderby'] ) )
		    $params['orderby'] = sanitize_key( $args['orderby'] );
		if ( isset( $args['order'] ) )
		    $params['order'] = sanitize_key( $args['order'] );
	    } elseif ( $args )
		return null;

	    $query = '';
	    if ( count( $params ) )
		$query = '?'.http_build_query( $params, '', '&' );

	    $result = $this->call_api('GET', '/v1/orders'.$query);

	    if ( isset( $result ) )
		return apply_filters( 'dyapress_premium_orders', $result, $args );
	}

	return null;
    }

    function create_order_payment( $order, $meta = false ) {
	// try to create a payment to fully pay the order
	// meta contains all the necessary data to process the payment
	// return a payment object
	if ( $this->is_active() ) {
	    $id = $this->get_order_id( $order );

	    if ( !empty( $id ) )
		$result = $this->call_api('POST', '/v1/orders/'.$id.'/payment', apply_filters( 'dyapress_premium_create_order_payment', $meta ) );

	    if ( isset( $result ) ) {
		// FIXME: avoid another API call to fetch order
		if ( !empty( $result->total_capturable ) && is_object( $order ) && empty( $order->total_capturable ) )
		    $order->total_capturable = $result->total_capturable;

		return apply_filters( 'dyapress_premium_payment_created', $result );
	    }
	}

	return null;
    }

    function create_invoices_payment( $invoices ) {
    // TODO: create_invoices_payment( invoices : { id : amount, ... }, ...)
    }

    function confirm_payment( $meta ) {
	if ( $this->is_active() ) {
	    if ( !empty( $meta ) )
		$result = $this->call_api('POST', '/v1/payments', apply_filters( 'dyapress_premium_confirm_payment', $meta ) );

	    if ( isset( $result ) )
		return apply_filters( 'dyapress_premium_payment_confirmed', $result );
	}

	return null;
    }
  
    // TODO: create_orders_payment( orders: { id : amount, ...}, ...) ???

    function is_uuid( $uuid ) {
	return $uuid == '' ||
	    preg_match('/^\{?[0-9a-f]{8}\-?[0-9a-f]{4}\-?[0-9a-f]{4}\-?[0-9a-f]{4}\-?[0-9a-f]{12}\}?$/i', $uuid) === 1;
    }

    function is_dolapikey( $key ) {
	return preg_match('/^[\w_]*$/', $key);
    }

    function is_vat_applicable( $thirdparty = null ) {
	if ( empty( $thirdparty ) )
	    $thirdparty = $this->get_current_thirdparty();

	if ( is_array( $thirdparty ) )
	    $thirdparty = (object) $thirdparty;

	if ( !empty( $thirdparty->country_id ) ) {
	    $country = $this->country();

	    if ( $country->id == $thirdparty->country_id )
		$result = true;
	    else {
		$european_countries = $this->get_european_countries();

		if ( isset( $european_countries[$country->id] ) && isset( $european_countries[$thirdparty->country_id] ) &&
		    !$this->check_vat_number( $thirdparty->vat_number, $thirdparty->country_id ) )
			$result = true;
		    else
			$result = false;
	    }
	} else
	    $result = true; // VAT is applicable by default if no country is given

	// allow VAT application correction (for example for sale of new car or boat in EU, ... which is not checked here
	return apply_filters( 'dyapress_premium_is_vat_applicable', $result, $thirdparty );
    }

    function check_vat_number( $vat_number, $country ) {
	if ( !isset( $vat_number) || $vat_number == '' )
	    return false;

	if ( isset( $country ) && (int)$country > 0 ) {
	    $countries = $this->get_countries();

	    if ( empty( $countries ) )
		return null;

	    // search for country code as country is id of get_countries()
	    foreach ( $countries as $c ) {
		if ( $c->id == (int)$country ) {
		    $country = $c->code;
		    break;
		}
	    }
	} elseif ( is_object( $country ) ) {
	    $country = $country->code;
	}

	// fix VAT number
	$vat_number = preg_replace( '/[^A-Z0-9]/', '', strtoupper( $vat_number ) );

	// assume country is now a country code and check VAT number
	if ( strlen( $country ) < 2 || substr( $vat_number, 0, strlen( $country ) ) != $country )
	    return false;

	// in case of France it is possible de check if VAT number is valid by computing the key directly
	if ( $country == 'FR' ) {
	    $key = substr( $vat_number, 2, 2 );
	    $siren = substr( $vat_number, 4 );

	    if ( $key != ( 12 + 3 * ( $siren % 97 ) ) % 97 )
		return false;
	}

	// if we are unable to check, simply return null
	if ( !class_exists( 'SoapClient' ) )
	    return null;

	// ready to call VIES webservice but this webservice is very expensive so check against transient before
	$result = get_transient( 'dyapress_premium_vat_number:'.$vat_number );
	if ( false === $result ) {
	    $client = new SoapClient( 'http://ec.europa.eu/taxation_customs/vies/checkVatService.wsdl' );
	    $params = array( 'countryCode' => $country, 'vatNumber' => substr( $vat_number, strlen( $country ) ) );
	    $resp = $client->checkVat( $params );

	    $result = $resp->valid ? 1 : 0;

	    // cache result in transient for seven days
	    set_transient( 'dyapress_premium_vat_number:'.$vat_number, $result, 7 * DAY_IN_SECONDS );
	}

	return $result;
    }
}
